# -*- coding: utf-8 -*-
"""Implements aligner for building STAR references."""

# pylint: disable=wildcard-import,redefined-builtin,unused-wildcard-import
from __future__ import absolute_import, division, print_function
from builtins import *
# pylint: enable=wildcard-import,redefined-builtin,unused-wildcard-import

import collections
import operator
import itertools
import sys
from typing import Any
import re

from intervaltree import IntervalTree
import numpy as np
import pandas as pd
import toolz

from imfusion.build.indexers.star import StarReference
from imfusion.external.star import star_align
from imfusion.external.stringtie import stringtie_assemble
from imfusion.external.compound import sort_bam
from imfusion.external.util import which, parse_arguments
from imfusion.model import Fusion, TransposonFusion
from imfusion.util import tabix, path

from .base import Aligner, register_aligner
from .. import util

CIGAR_MATCH_REGEX = re.compile(r'(\d+)M')


class StarAligner(Aligner):
    """STAR aligner.

    Aligner that identifies insertions using STAR, by aligning reads using STAR
    (including chimeric alignments) and identifying gene-transposon fusions
    from chimeric read alignments.

    As STAR provides chimeric alignment information but does not provide a
    an actual list of fusions, we identify fusions by summarizing these
    alignments ourselves. The strategy to do so is as follows:

        1. We filter for chimeric reads involving the transposon and genomic
           loci, to reduce the number of reads we need to consider.
        2. We select chimeric reads (or mates) that actually overlap the
           breakpoint of the fusion, as these alignments provide the exact
           position of the fusion. These reads are grouped by their positions
           to determine the number of reads that support each fusion. Fusions
           with close distance (on both the genomic and transposon loci) are
           merged to account for slight variations in the alignment.
        3. For paired-end data:
            - We select mates that span the fusion site
              (meaning that they do not overlap the breakpoint, therefore we
              only have an approximate position) and group mates of which both
              ends are with close proximity of eachother.
            - We then try to assign these groups of spanning mates to
              previously identified 'junction' fusions that are in the close
              vicinity. If we find a match, the number of mates in the group is
              registered as 'spanning support' for that fusion. If no match is
              found, we consider the group to identify a unique fusion and
              derive an approximate location for the fusion from the
              mate locations.
        4. We return the combined list of fusions. For single-end data, this
           only contains junction fusions, whilst for paired-end data this is
           the combined list of junction and unassigned spanning fusions.

    To avoid memory issues with STAR during the sorting of bam files, sorting
    can be performed using Sambamba. Other external dependencies of this
    aligner are STAR itself and Stringtie, which is used for the assembly of
    novel transcripts (if requested).

    Parameters
    ----------
    reference : Reference
        Reference class describing the reference paths.
    assemble : bool
        Whether to perform the assembly of novel transcripts using Stringie.
    assemble_args : Dict[str, Any]
        Extra arguments to pass to Stringie during assembly.
    min_flank : int
        Minimum number of flanking bases required on both sides of the
        breakpoint for a fusion to be considered valid. This means that, for
        a split read overlapping the fusion boundary, at least ``min_flank``
        bases should be on either sides of the fusion.
    threads : int
        Number of threads to use.
    extra_args :
        Extra arguments to pass to STAR for alignment.
    logger : logging.Logger
        Logger to be used for logging messages.
    filter_features : bool
        Whether insertions should be filtered to remove non SA/SD insertions.
    filter_orientation : bool
        Whether insertions should be filtered for fusions in which the
        transposon feature is in the wrong orientation.
    filter_blacklist : List[str]
        List of gene ids to filter insertions for.
    external_sort : bool
        Whether sorting should be performed using STAR (False) or using
        Sambamba (True). Sambamba uses less memory than STAR, but is slower.
    merge_junction_dist : int
        Maximum distance within which fusions supported by split reads
        (overlapping the junction, so that the exact breakpoint is known) are
        merged. This merging avoids calling multiple fusions due to slight
        variations in the alignment, although this value should not be chosen
        too large to avoid merging distinct insertions.
    max_spanning_dist : int
        Maximum distance within which spanning mate pairs (mates that do not
        overlap the fusion junction) are grouped when summarizing spanning
        chimeric reads. Both mates from two pairs need to be within this
        distance of each other to be merged. The value should be chosen to
        reflect the expected or emprical insert size.
    max_junction_dist : int
        Maxixmum distance within which groups of spanning mates are assigned
        to a junction fusion (which is supported by split reads, so that its
        exact position is known). Groups that cannot be assigned to a junction
        fusion are considered to arise from a separate insertion.

    """

    def __init__(
            self,
            reference,  # type: StarReference
            assemble=False,  # type: bool
            assemble_args=None,  # type: Dict[str, Any]
            min_flank=12,  # type: int
            threads=1,  # type: int
            extra_args=None,  # type: Dict[str, Any]
            logger=None,  # type: Any
            filter_features=True,  # type: bool
            filter_orientation=True,  # type: bool
            filter_blacklist=None,  # type: List[str]
            external_sort=False,  # type: bool
            merge_junction_dist=10,  # type: int
            max_spanning_dist=300,  # type: int
            max_junction_dist=10000  # type: int
    ):  # type: (...) -> None

        super().__init__(reference=reference, logger=logger)

        self._assemble = assemble
        self._assemble_args = assemble_args or {}
        self._min_flank = min_flank
        self._threads = threads
        self._external_sort = external_sort
        self._extra_args = extra_args or {}

        self._merge_junction_dist = merge_junction_dist
        self._max_spanning_dist = max_spanning_dist
        self._max_junction_dist = max_junction_dist

        self._filter_features = filter_features
        self._filter_orientation = filter_orientation
        self._filter_blacklist = filter_blacklist

    @property
    def dependencies(self):
        """External dependencies required by aligner."""

        programs = ['STAR']

        if self._assemble:
            programs += ['stringtie']

        return programs

    def identify_insertions(self, fastq_path, output_dir, fastq2_path=None):
        """Identifies insertions from given reads.

        Aligns RNA-seq reads to the reference genome and uses this alignment
        to identify gene-transposon fusions. These gene-transposon fusions
        are summarized to transposon insertions and returned.

        Parameters
        ----------
        fastq_path : Path
            Path to fastq file containing sequence reads. For paired-end data,
            this should refer to the first read of the pair.
        output_dir : Path
            Output directory, to use for output files such as the alignment.
        fastq2_path : Path
            For paired-end sequencing data, path to fastq file containing
            the second read of the pair.

        Yields
        ------
        Insertion
            Yields the identified insertions.
        """

        output_dir.mkdir(parents=True, exist_ok=True)

        # Perform alignment using STAR.
        alignment_path = output_dir / 'alignment.bam'
        if not alignment_path.exists():
            self._logger.info('Performing alignment using STAR')
            self._align(fastq_path, output_dir, fastq2_path=fastq2_path)
        else:
            self._logger.info('Using existing STAR alignment')

        # Assemble transcripts if requested.
        if self._assemble:
            assembled_path = output_dir / 'assembled.gtf.gz'
            if not assembled_path.exists():
                self._logger.info('Assembling transcripts using Stringtie')

                # Generate assembled GTF.
                stringtie_out_path = assembled_path.with_suffix('')
                stringtie_assemble(
                    alignment_path,
                    gtf_path=self._reference.gtf_path,
                    output_path=stringtie_out_path)

                # Compress and index.
                tabix.index_gtf(stringtie_out_path, output_path=assembled_path)
                stringtie_out_path.unlink()
            else:
                self._logger.info('Using existing Stringtie assembly')
        else:
            assembled_path = None

        # Extract identified fusions.
        self._logger.info('Extracting gene-transposon fusions')
        fusion_path = output_dir / 'fusions.out'
        fusions = list(self._extract_fusions(fusion_path))

        # Extract insertions.
        self._logger.info('Summarizing insertions')
        insertions = list(
            util.extract_insertions(
                fusions,
                gtf_path=self._reference.indexed_gtf_path,
                features_path=self._reference.features_path,
                assembled_gtf_path=assembled_path,
                ffpm_fastq_path=fastq_path,
                chromosomes=None))

        # self._logger.info('Filtering insertions')
        insertions = util.filter_insertions(
            insertions,
            features=self._filter_features,
            orientation=self._filter_orientation,
            blacklist=self._filter_blacklist)

        for insertion in insertions:
            yield insertion

    def _align(self, fastq_path, output_dir, fastq2_path=None):
        # Put output into subdirectory, we will symlink the
        # expected outputs from STAR later.
        star_dir = output_dir / '_star'

        # Gather default arguments.
        sort_type = 'Unsorted' if self._external_sort else 'SortedByCoordinate'

        args = {
            '--twopassMode': ('Basic', ),
            '--outReadsUnmapped': ('None', ),
            '--outSAMtype': ('BAM', sort_type),
            '--runThreadN': (self._threads, ),
            '--chimSegmentMin': (self._min_flank, ),
            '--outSAMstrandField': ('intronMotif', )  # XS field for Stringtie.
        }

        star_align(
            fastq_path=fastq_path,
            fastq2_path=fastq2_path,
            index_path=self._reference.index_path,
            output_dir=star_dir,
            extra_args=toolz.merge(args, self._extra_args))

        # If not yet sorted, sort bam file using samtools/sambamba.
        sorted_bam_path = star_dir / 'Aligned.sortedByCoord.out.bam'

        if sort_type == 'Unsorted':
            unsorted_bam_path = star_dir / 'Aligned.out.bam'
            sort_bam(unsorted_bam_path, sorted_bam_path, threads=self._threads)
            unsorted_bam_path.unlink()

        # Symlink fusions and alignment into expected location.
        dest_bam_path = output_dir / 'alignment.bam'
        path.symlink_relative(
            src_path=sorted_bam_path, dest_path=dest_bam_path)

        path.symlink_relative(
            src_path=star_dir / 'Chimeric.out.junction',
            dest_path=output_dir / 'fusions.out')

    def _extract_fusions(self, fusion_path):
        # Read chimeric junction data.
        chimeric_data = read_chimeric_junctions(fusion_path)

        # Extract transposon fusions.
        fusions = extract_transposon_fusions(
            chimeric_data,
            self._reference.transposon_name,
            merge_junction_dist=self._merge_junction_dist,
            max_spanning_dist=self._max_spanning_dist,
            max_junction_dist=self._max_junction_dist)

        for fusion in fusions:
            yield fusion

    @classmethod
    def configure_args(cls, parser):
        """Configures an argument parser for the Indexer.

        Used by ``imfusion-build`` to configure the sub-command for
        this indexer (if registered as an Indexer using the
        ``register_indexer`` function).

        Parameters
        ----------
        parser : argparse.ArgumentParser
            Argument parser to configure.

        """

        super().configure_args(parser)

        star_group = parser.add_argument_group('STAR arguments')
        star_group.add_argument('--star_threads', type=int, default=1)
        star_group.add_argument('--star_min_flank', type=int, default=12)
        star_group.add_argument(
            '--star_external_sort', default=False, action='store_true')
        star_group.add_argument('--star_args', default='')

        star_group.add_argument('--merge_junction_dist', default=10, type=int)
        star_group.add_argument('--max_spanning_dist', default=300, type=int)
        star_group.add_argument('--max_junction_dist', default=10000, type=int)

        assemble_group = parser.add_argument_group('Assembly')
        assemble_group.add_argument(
            '--assemble', default=False, action='store_true')

        filt_group = parser.add_argument_group('Filtering')
        filt_group.add_argument(
            '--no_filter_orientation',
            dest='filter_orientation',
            default=True,
            action='store_false')
        filt_group.add_argument(
            '--no_filter_feature',
            dest='filter_features',
            default=True,
            action='store_false')
        filt_group.add_argument('--blacklisted_genes', nargs='+')

    @classmethod
    def _parse_args(cls, args):
        kws = dict(
            reference=StarReference(args.reference),
            min_flank=args.star_min_flank,
            threads=args.star_threads,
            extra_args=parse_arguments(args.star_args),
            external_sort=args.star_external_sort,
            assemble=args.assemble,
            merge_junction_dist=args.merge_junction_dist,
            max_spanning_dist=args.max_spanning_dist,
            max_junction_dist=args.max_junction_dist,
            filter_features=args.filter_features,
            filter_orientation=args.filter_orientation,
            filter_blacklist=args.blacklisted_genes)

        return toolz.merge(super()._parse_args(args), kws)


register_aligner('star', StarAligner)


def read_chimeric_junctions(chimeric_path):
    """Reads junctions from STARs Chimeric.out.junction output file.

    Parameters
    ----------
    pathlib.Path
        Path to the Chimeric.out.junction file.

    Returns
    -------
    pd.DataFrame
        Pandas DataFrame containing the junction data.

    """

    # File structure is as follows.
    #
    # http://labshare.cshl.edu/shares/gingeraslab/www-data/
    # dobin/STAR/STAR.posix/doc/STARmanual.pdf
    #
    # The first 9 columns give information about the chimeric junction:
    #
    #   column 1: chromosome of the donor
    #   column 2: first base of the intron of the donor (1-based)
    #   column 3: strand of the donor
    #   column 4: chromosome of the acceptor
    #   column 5: first base of the intron of the acceptor (1-based)
    #   column 6: strand of the acceptor
    #   column 7: junction type: -1=encompassing junction
    #       (between the mates), 1=GT/AG, 2=CT/AC
    #   column 8: repeat length to the left of the junction
    #   column 9: repeat length to the right of the junction
    #
    # Columns 10-14 describe the alignments of the two chimeric segments,
    # it is SAM like. Alignments are given with respect to the (+) strand
    #
    #   column 10: read name
    #   column 11: first base of the first segment (on the + strand)
    #   column 12: CIGAR of the first segment
    #   column 13: first base of the second segment
    #   column 14: CIGAR of the second segment
    #

    names = [
        'seqname_a', 'location_a', 'strand_a', 'seqname_b', 'location_b',
        'strand_b', 'junction_type', 'repeat_length_left',
        'repeat_length_right', 'read_name', 'first_segment_base',
        'first_segment_cigar', 'second_segment_base', 'second_segment_cigar'
    ]

    junctions = pd.read_csv(
        str(chimeric_path),
        sep='\t',
        header=None,
        names=names,
        dtype={'seqname_a': 'str',
               'seqname_b': 'str'})

    for strand_col in ['strand_a', 'strand_b']:
        junctions[strand_col] = junctions[strand_col].map({'+': 1, '-': -1})

    return junctions


def normalize_chimeric_junctions(chimeric_data, seqname=None):
    """Normalizes chimeric junction data so that seqnames are ordered.

    This function normalizes a chimeric junction DataFrame so that seqname_a
    and seqname_b are lexically ordered (i.e., seqname_a < seqname_b).
    Intra-chromosomal junctions (with seqname_a == seqname_b) are normalized
    using their location, so that location_a < location_b.

    Alternatively, if the seqname argument is given, the junctions are
    normalized so that seqname_a == seqname. Any junctions not including
    the seqname chromosome are dropped.

    Parameters
    ----------
    chimeric_data : pd.DataFrame
        DataFrame of chimeric junctions to normalize.
    seqname : str, optional
        Optional sequence name to normalize for. If given, junctions are
        normalized so that seqname_a == seqname. Any junctions not including
        the named sequence are dropped.

    Returns
    -------
    pd.DataFrame
        Normalized chimeric junction DataFrame.

    """

    if seqname is None:
        # Reads are 'normal' if seq_a < seq_b or, in the case that
        # seq_a == seq_b, if loc_a < loc_b.
        same_seq = chimeric_data.seqname_a == chimeric_data.seqname_b
        seq_ordered = (chimeric_data.seqname_a < chimeric_data.seqname_b)
        loc_ordered = (chimeric_data.location_a < chimeric_data.location_b)

        norm_mask = (seq_ordered | (same_seq & loc_ordered))
    else:
        # Subset to samples that have at least one end on seqname.
        chimeric_data = chimeric_data.ix[(chimeric_data.seqname_a == seqname) |
                                         (chimeric_data.seqname_b == seqname)]

        # Chimeric reads are 'normal' if first side is on seqname and
        # and if locations are ordered in the case that seq_a == seq_b.
        loc_ordered = (chimeric_data.location_a < chimeric_data.location_b)
        norm_mask = (chimeric_data.seqname_a == seqname) & loc_ordered

    normed = pd.concat(
        [
            chimeric_data.ix[norm_mask],
            _reverse_chimeric(chimeric_data.ix[~norm_mask])
        ],
        axis=0)

    return normed


def _reverse_chimeric(chimeric_data):
    return (chimeric_data.rename(columns={
        'seqname_a': 'seqname_b',
        'location_a': 'location_b',
        'strand_a': 'strand_b',
        'seqname_b': 'seqname_a',
        'location_b': 'location_a',
        'strand_b': 'strand_a',
        'flank_a': 'flank_b',
        'flank_b': 'flank_a'
    }).assign(
        strand_a=lambda df: df.strand_a * -1,
        strand_b=lambda df: df.strand_b * -1).get(chimeric_data.columns))


def extract_transposon_fusions(chimeric_data,
                               transposon_name,
                               merge_junction_dist=10,
                               max_spanning_dist=300,
                               max_junction_dist=10000):
    """Extracts transposon fusions from a STAR chimeric read dataframe."""

    # Subset and normalize chimeric reads for transposon.
    chimeric_data = chimeric_data.ix[
        (chimeric_data['seqname_a'] == transposon_name) ^
        (chimeric_data['seqname_b'] == transposon_name)] # yapf: disable

    chimeric_data = normalize_chimeric_junctions(
        chimeric_data, seqname=transposon_name)

    # Extract junction fusions and merge close junctions.
    junctions = list(
        extract_junction_fusions(
            chimeric_data, merge_dist=merge_junction_dist))

    # Assign spanning reads to junctions.
    junctions, unassigned = assign_spanning_reads(
        junctions,
        chimeric_data,
        max_dist_left=max_spanning_dist,
        max_dist_right=max_junction_dist)

    # Extract spanning fusions from unused reads.
    spanning = extract_spanning_fusions(unassigned, max_dist=max_spanning_dist)
    fusions = itertools.chain.from_iterable([junctions, spanning])

    # Convert to transposon fusions.
    for fusion in fusions:
        yield TransposonFusion.from_fusion(fusion, transposon_name)


def extract_junction_fusions(chimeric_data, merge_dist=None):
    """Extracts junction fusions from a STAR chimeric read dataframe."""

    # Ensure chimeric data only contains junction reads.
    chimeric_data = chimeric_data.query('junction_type >= 0')

    if len(chimeric_data) > 0:
        # Add flanking annotation.
        chimeric_data = chimeric_data.copy()

        flanks = (_flank_sizes(t) for t in chimeric_data.itertuples())
        chimeric_data['flank_a'], chimeric_data['flank_b'] = zip(*flanks)

        # Group by position and summarize.
        grouped = chimeric_data.groupby([
            'seqname_a', 'location_a', 'strand_a', 'seqname_b', 'location_b',
            'strand_b'
        ])

        summarized = (grouped.agg({
            'flank_a': 'max',
            'flank_b': 'max',
            'read_name': 'nunique'
        }).reset_index().assign(support_spanning=0)
                      .rename(columns={'read_name': 'support_junction'}))

        # Transform to Fusions.
        fusions = (Fusion(**toolz.keyfilter(lambda k: k not in {'Index'},
                                            row._asdict()))
                   for row in summarized.itertuples())

        # Merge fusions within dist.
        if merge_dist is not None:
            fusions = Fusion.merge(fusions, max_dist=merge_dist)

        for fusion in fusions:
            yield fusion


def _flank_sizes(row):
    return (_flank_size(row, donor=True), _flank_size(row, donor=False))


def _flank_size(row, donor=True):
    """Calculates flank size of a spanning read on one side of the fusion."""

    if donor:
        mate_index = -1 if row.strand_a == 1 else 0
        cigar = row.first_segment_cigar
    else:
        mate_index = 0 if row.strand_b == 1 else -1
        cigar = row.second_segment_cigar

    cigar_split = cigar.split('p')
    size = _sum_matches(cigar_split[mate_index])

    return size


def _sum_matches(cigar_str):
    """Sums the number of matches in a CIGAR string."""
    matches = CIGAR_MATCH_REGEX.findall(cigar_str)
    return sum(map(int, matches))


def assign_spanning_reads(junctions, chimeric_data, max_dist_left,
                          max_dist_right):
    """Assigns spanning reads to the closest junction that they support."""

    # Ensure chimeric data only contains spanning reads.
    chimeric_data = chimeric_data.query('junction_type < 0')

    # Assign chimeric reads to junctions.
    trees_a = _build_lookup_trees(junctions, side='a')
    trees_b = _build_lookup_trees(junctions, side='b')

    assignments = collections.defaultdict(list)
    unassigned_idx = []

    for spanning_read in chimeric_data.itertuples():
        junc = _lookup_closest(spanning_read, trees_a, trees_b, max_dist_left,
                               max_dist_right)

        if junc is not None:
            assignments[junc].append(spanning_read)
        else:
            unassigned_idx.append(spanning_read.Index)

    # Augment junctions.
    new_juncs = [
        junc._replace(
            support_spanning=junc.support_spanning + len(assignments[junc]))
        for junc in junctions
    ]

    # Select unassigned reads.
    unassigned = chimeric_data.ix[unassigned_idx]

    return new_juncs, unassigned


def _build_lookup_trees(fusions, side='a'):
    # Define attr getters.
    get_seq = operator.attrgetter('seqname_' + side)
    get_loc = operator.attrgetter('location_' + side)
    get_strand = operator.attrgetter('strand_' + side)

    # Setup keyfunc.
    def _keyfunc(fusion):
        return (get_seq(fusion), get_strand(fusion))

    # Sort and group fusions.
    fusions = sorted(fusions, key=_keyfunc)
    grouped = itertools.groupby(fusions, key=_keyfunc)

    trees = {}
    for key, grp in grouped:
        trees[key] = IntervalTree.from_tuples(
            (get_loc(fus), get_loc(fus) + 1, fus) for fus in grp)

    return trees


def _lookup_closest(fusion,
                    trees_a,
                    trees_b,
                    max_dist_left,
                    max_dist_right,
                    slack=5):
    def _dist(fusion, hit):
        return (abs(fusion.location_a - hit.location_a) +
                abs(fusion.location_b - hit.location_b))

    # Identify candidate fusions.
    overlap_a = _lookup_tree(
        fusion, trees_a, side='a', max_dist=max_dist_left, slack=slack)
    overlap_b = _lookup_tree(
        fusion, trees_b, side='b', max_dist=max_dist_right, slack=slack)
    overlap = list(overlap_a & overlap_b)

    if len(overlap) > 0:
        distances = np.array(_dist(fusion, hit) for hit in overlap)
        return overlap[distances.argmin()]
    else:
        return None


def _lookup_tree(fusion, trees, side, max_dist=300, slack=5):

    # Define attr getters.
    get_seq = operator.attrgetter('seqname_' + side)
    get_loc = operator.attrgetter('location_' + side)
    get_strand = operator.attrgetter('strand_' + side)

    # Setup tree, query range.
    strand = get_strand(fusion)

    try:
        tree = trees[(get_seq(fusion), strand)]

        direction = (-1 if side == 'a' else 1) * strand

        loc = get_loc(fusion)
        if direction == 1:
            start, end = loc - max_dist, loc + slack
        else:
            start, end = loc - slack, loc + max_dist

        # Determine overlap and return fusions.
        overlap = set(tup[2] for tup in tree[start:end])

        return overlap
    except KeyError:
        return set()


def extract_spanning_fusions(chimeric_data, max_dist):
    """Extracts spanning fusions from a STAR chimeric read dataframe."""

    # Select spanning fusions.
    chimeric_data = chimeric_data.query('junction_type < 0')

    # Group by position and summarize.
    for grp in _groupby_position(chimeric_data, max_dist=max_dist):
        first = grp.iloc[0]

        agg_donor = max if first.strand_a == 1 else min
        agg_acc = min if first.strand_b == 1 else max

        yield Fusion(
            seqname_a=first.seqname_a,
            location_a=agg_donor(grp.location_a),
            strand_a=first.strand_a,
            seqname_b=first.seqname_b,
            location_b=agg_acc(grp.location_b),
            strand_b=first.strand_b,
            support_junction=0,
            support_spanning=len(grp),
            flank_a=0,
            flank_b=0)


def _groupby_position(spanning_data, max_dist):
    """Groups reads within given distance at both locations."""

    grp_cols = ['seqname_a', 'seqname_b', 'strand_a', 'strand_b']

    for _, grp in spanning_data.groupby(grp_cols):
        if len(grp) == 1:
            yield grp
        else:
            clust_a = _cluster_positions(
                grp, col='location_a', max_dist=max_dist)
            clust_b = _cluster_positions(
                grp, col='location_b', max_dist=max_dist)

            for _, loc_grp in grp.groupby([clust_a, clust_b]):
                yield loc_grp


def _cluster_positions(df, col, max_dist):
    locs = df[col].sort_values()

    diff = np.diff(np.insert(locs.values, 0, 0))
    groups = pd.Series(np.cumsum(diff > max_dist), index=locs.index)

    return groups.ix[df.index]



"""Module containing functions for calling STAR."""

from .util import flatten_arguments, run_command


def star_index(fasta_path,
               gtf_path,
               output_dir,
               overhang=100,
               threads=1,
               log_path=None):
    """Builds a bowtie index for a reference genome.

    Parameters
    ----------
    fasta_path : pathlib.Path
        Path to the reference genome.
    gtf_path : pathlib.Path
        Path to the reference gtf file.
    output_dir pathlib.Path:
        Output directory for the built index.
    overhang : int
        Value to use for sjdbOverhang (see STAR manual for more details).
    threads : int
        Number of threads to use.
    log_path : pathlib.Path
        Where to write the log output.

    """

    output_dir.mkdir(exist_ok=False, parents=True)

    args = [
        'STAR', '--runMode', 'genomeGenerate', '--genomeDir', str(output_dir),
        '--genomeFastaFiles', str(fasta_path), '--sjdbGTFfile', str(gtf_path),
        '--sjdbOverhang', str(overhang), '--runThreadN', str(threads)
    ]

    run_command(args=args, log_path=log_path)


def star_align(
        fastq_path,
        index_path,
        output_dir,
        fastq2_path=None,
        extra_args=None,
        threads=1,
        log_path=None, ):
    """Runs STAR in alignment mode for the given fastqs."""

    extra_args = extra_args or {}

    if not output_dir.exists():
        output_dir.mkdir(parents=True)

    # Check if files are gzipped.
    if fastq_path.suffixes[-1] == '.gz':
        extra_args['--readFilesCommand'] = ('gunzip', '-c')

    if threads > 1:
        extra_args['--runThreadN'] = threads

    # Assemble arguments.
    if fastq2_path is None:
        fastq_args = [str(fastq_path)]
    else:
        fastq_args = [str(fastq_path), str(fastq2_path)]

    args = [
        'STAR', '--genomeDir', str(index_path), '--outFileNamePrefix',
        str(output_dir) + '/', '--readFilesIn'
    ] + fastq_args

    args += [str(arg) for arg in flatten_arguments(extra_args or {})]

    run_command(args=args, log_path=log_path)



# -*- coding: utf-8 -*-
"""Unit tests for functions calling sambamba externally."""

from pathlib2 import Path

import pytest

from imfusion.external import sambamba


class TestSambambaSort(object):
    """Unit tests for the sambamba_sort function."""

    def test_call(self, mocker):
        """Tests example call with Path paths."""

        mock_run = mocker.patch.object(sambamba, 'run_command')
        sambamba.sambamba_sort(Path('test.bam'), Path('sorted.bam'))

        mock_run.assert_called_once_with(
            args=[
                'sambamba', 'sort', '-o', 'sorted.bam', '--tmpdir=_tmp', '-t',
                '1', 'test.bam'
            ],
            log_path=None)



# -*- coding: utf-8 -*-
"""Module containing model classes for fusions and insertions."""

# pylint: disable=wildcard-import,redefined-builtin,unused-wildcard-import
from __future__ import absolute_import, division, print_function
from builtins import *
# pylint: enable=wildcard-import,redefined-builtin,unused-wildcard-import

import collections
import itertools
import operator

import numpy as np
import pandas as pd
import toolz

from imfusion.util.frozendict import frozendict


class FrameMixin(object):
    @classmethod
    def _get_columns(cls):
        return cls._fields

    @classmethod
    def _to_record(cls, obj):
        return obj._asdict()

    @classmethod
    def _to_obj(cls, record):
        return cls(**record._asdict())

    @classmethod
    def read_csv(cls, file_path, **kwargs):
        """Reads objects from a csv file into a pandas DataFrame."""
        df = pd.read_csv(str(file_path), **kwargs)
        return cls.format_frame(df)

    @classmethod
    def to_csv(cls, file_path, objects, index=False, **kwargs):
        """Writes objects to a csv file."""
        df = cls.to_frame(objects)
        df.to_csv(str(file_path), index=index, **kwargs)

    @classmethod
    def from_csv(cls, file_path, **kwargs):
        """Reads objects from a csv file."""

        df = cls.read_csv(file_path, **kwargs)

        for obj in cls.from_frame(df):
            yield obj

    @classmethod
    def from_frame(cls, df):
        """Converts dataframe into an interable of objects."""
        for tup in df.itertuples():
            yield cls._to_obj(tup)

    @classmethod
    def format_frame(cls, df):
        """Formats dataframe into a cohorent format."""
        return cls._reorder_columns(df, order=cls._get_columns())

    @classmethod
    def to_frame(cls, objects):
        """Converts list of objects to a dataframe representation."""

        # Check if insertions is empty.
        is_empty, objects = cls._is_empty(objects)

        if is_empty:
            df = pd.DataFrame.from_records([], columns=cls._get_columns())
        else:
            rows = (cls._to_record(obj) for obj in objects)
            df = pd.DataFrame.from_records(rows)
            df = cls.format_frame(df)

        return df

    @staticmethod
    def _is_empty(iterable):
        try:
            _, iterable = toolz.peek(iterable)
            empty = False
        except StopIteration:
            empty = True

        return empty, iterable

    @classmethod
    def _reorder_columns(cls, df, order):
        extra_cols = set(df.columns) - set(order)
        col_order = list(order) + sorted(extra_cols)
        return df[col_order]


class MetadataFrameMixin(FrameMixin):
    """Mixin class adding namedtuple/frame conversion support."""

    @classmethod
    def _get_columns(cls):
        fields = list(cls._fields)
        del fields[fields.index('metadata')]
        return fields

    @classmethod
    def _to_record(cls, obj):
        obj_data = obj._asdict()
        metadata = obj_data.pop('metadata')
        return toolz.merge(metadata, obj_data)

    @classmethod
    def _to_obj(cls, record):
        record_dict = record._asdict()

        metadata_fields = [
            k for k in record_dict.keys() if k not in set(cls._get_columns())
        ]
        metadata = {k: record_dict.pop(k) for k in metadata_fields}

        metadata.pop('Index', None)

        return cls(metadata=frozendict(metadata), **record_dict)

    def __getattr__(self, name):
        if name in self.metadata:
            return self.metadata[name]
        else:
            raise AttributeError


_Fusion = collections.namedtuple('Fusion', [
    'seqname_a', 'location_a', 'strand_a', 'seqname_b', 'location_b',
    'strand_b', 'flank_a', 'flank_b', 'support_junction', 'support_spanning'
])


class Fusion(FrameMixin, _Fusion):
    """Model Fusion class."""

    __slots__ = ()

    @property
    def support(self):
        """Support score for the fusion."""
        return self.support_junction + self.support_spanning

    def normalize(self, seqname=None):
        """Normalizes fusions so that the side whose seqname has the
           lowest lexical ordering is used as donor.
        """

        if seqname is not None:
            if self.seqname_a == seqname:
                is_norm = True
            else:
                if self.seqname_b != seqname:
                    raise ValueError('Fusion does not include given seqname')
                is_norm = False
        elif self.seqname_a != self.seqname_b:
            is_norm = self.seqname_a < self.seqname_b
        else:
            is_norm = self.location_a < self.location_b

        if is_norm:
            return self
        else:
            return self._replace(
                seqname_a=self.seqname_b,
                location_a=self.location_b,
                strand_a=self.strand_b * -1,
                seqname_b=self.seqname_a,
                location_b=self.location_a,
                strand_b=self.strand_a * -1,
                flank_a=self.flank_b,
                flank_b=self.flank_a)

    def distance(self, other):
        """Determine distance to other fusion."""

        if (self.seqname_a != other.seqname_a or
                self.seqname_b != other.seqname_b):
            raise ValueError('Fusions are on different reference sequences')

        return (abs(self.location_a - other.location_a) +
                abs(self.location_b - other.location_b))

    @classmethod
    def merge(cls, junctions, max_dist):
        """Merges fusion within genomic distance into a single fusion."""
        for grp in cls._groupby_genomic_position(junctions, max_dist):
            if len(grp) == 1:
                yield grp[0]
            else:
                get_support = operator.attrgetter('support_junction')
                sorted_grp = sorted(grp, key=get_support)

                yield sorted_grp[-1]._replace(
                    support_junction=sum(f.support_junction
                                         for f in sorted_grp),
                    support_spanning=sum(f.support_spanning
                                         for f in sorted_grp),
                    flank_a=max(f.flank_a for f in sorted_grp),
                    flank_b=max(f.flank_b for f in sorted_grp))

    @classmethod
    def _groupby_genomic_position(cls, junctions, max_dist):

        # Group junctions by strand and sequence.
        def _keyfunc(fusion):
            return (fusion.seqname_a, fusion.strand_a, fusion.seqname_b,
                    fusion.strand_b)

        sorted_juncs = sorted(junctions, key=_keyfunc)
        grouped_juncs = itertools.groupby(sorted_juncs, key=_keyfunc)

        for _, grp in grouped_juncs:
            for grp_a in cls._groupby_position(grp, 'a', max_dist):
                for grp_b in cls._groupby_position(grp_a, 'b', max_dist):
                    yield grp_b

    @staticmethod
    def _groupby_position(junctions, side, max_dist):
        get_loc = operator.attrgetter('location_' + side)
        sorted_juncs = sorted(junctions, key=get_loc)

        grp, prev_pos = [], np.nan
        for junc in sorted_juncs:
            if (get_loc(junc) - prev_pos) > max_dist:
                yield grp
                grp = [junc]
            else:
                grp.append(junc)
            prev_pos = get_loc(junc)

        if len(grp) > 0:
            yield grp


_TransposonFusion = collections.namedtuple('TransposonFusion', [
    'seqname', 'anchor_genome', 'anchor_transposon', 'strand_genome',
    'strand_transposon', 'flank_genome', 'flank_transposon',
    'support_junction', 'support_spanning', 'metadata'
])


class TransposonFusion(MetadataFrameMixin, _TransposonFusion):
    """Model class representing a gene-transposon fusion."""

    __slots__ = ()

    @property
    def support(self):
        """Support score for the fusion."""
        return self.support_junction + self.support_spanning

    @property
    def genome_region(self):
        """Returns the spanned region on the genome."""
        if self.flank_genome > 0:
            start = self.anchor_genome
            end = self.anchor_genome + self.flank_genome
        else:
            start = self.anchor_genome + self.flank_genome
            end = self.anchor_genome
        return self.seqname, start, end

    @property
    def transposon_region(self):
        """Returns the spanned region on the transposon."""
        if self.flank_transposon > 0:
            start = self.anchor_transposon
            end = self.anchor_transposon + self.flank_transposon
        else:
            start = self.anchor_transposon + self.flank_transposon
            end = self.anchor_transposon
        return start, end

    @classmethod
    def from_fusion(cls, fusion, transposon_name, metadata=None):
        """Converts the fusion to a transposon fusion object."""

        if (fusion.seqname_a == transposon_name and
                fusion.seqname_b == transposon_name):
            raise ValueError('Fusion does not involve genomic sequence')
        elif (fusion.seqname_a != transposon_name and
              fusion.seqname_b != transposon_name):
            raise ValueError('Fusion does not involve transposon')
        elif fusion.seqname_a == transposon_name:
            tr_key, gen_key = 'a', 'b'
            tr_flank = fusion.flank_a * -fusion.strand_a
            gen_flank = fusion.flank_b * fusion.strand_b
        else:
            tr_key, gen_key = 'b', 'a'
            tr_flank = fusion.flank_b * fusion.strand_b
            gen_flank = fusion.flank_a * -fusion.strand_a

        return TransposonFusion(
            seqname=getattr(fusion, 'seqname_' + gen_key),
            anchor_genome=getattr(fusion, 'location_' + gen_key),
            anchor_transposon=getattr(fusion, 'location_' + tr_key),
            strand_genome=getattr(fusion, 'strand_' + gen_key),
            strand_transposon=getattr(fusion, 'strand_' + tr_key),
            flank_genome=gen_flank,
            flank_transposon=tr_flank,
            support_junction=fusion.support_junction,
            support_spanning=fusion.support_spanning,
            metadata=frozendict(metadata or {}))


_Insertion = collections.namedtuple('Insertion', [
    'id', 'seqname', 'position', 'strand', 'support_junction',
    'support_spanning', 'support', 'metadata'
])


class Insertion(MetadataFrameMixin, _Insertion):
    """Model class representing an insertion."""

    __slots__ = ()

    @classmethod
    def from_transposon_fusion(cls, fusion, id_=None, drop_metadata=None):
        """Converts (annotated) transposon fusion to an insertion.

        Requires

        Parameters
        ----------
        id : str
            ID to use for the insertion.

        Returns
        -------
        Insertion
            Insertion object derived from the transposon fusion.

        """

        if drop_metadata is None:
            drop_metadata = {'strand_genome', 'strand_transposon'}

        # Fusion should have annotated transposon feature.
        if 'feature_name' not in fusion.metadata:
            raise ValueError(
                'Fusion does not have an annotated transposon feature')

        strand = fusion.strand_genome * fusion.strand_transposon

        if 'gene_strand' in fusion.metadata:
            gene_strand = fusion.metadata['gene_strand']
            orientation = 'sense' if strand == gene_strand else 'antisense'
        else:
            orientation = None

        ins_metadata = toolz.keyfilter(lambda k: k not in drop_metadata,
                                       fusion.metadata)
        ins_metadata['transposon_anchor'] = fusion.anchor_transposon

        if orientation is not None:
            ins_metadata['orientation'] = orientation

        return Insertion(
            id=id_,
            seqname=fusion.seqname,
            position=fusion.anchor_genome,
            strand=strand,
            support_junction=fusion.support_junction,
            support_spanning=fusion.support_spanning,
            support=fusion.support,
            metadata=frozendict(ins_metadata))

    @classmethod
    def from_transposon_fusions(cls,
                                fusions,
                                id_fmt_str=None,
                                drop_metadata=None):
        """Converts annotated transposon fusions to insertions."""

        if id_fmt_str is not None:
            insertions = (cls.from_transposon_fusion(
                fusion,
                id_=id_fmt_str.format(i + 1),
                drop_metadata=drop_metadata)
                          for i, fusion in enumerate(fusions))
        else:
            insertions = (cls.from_transposon_fusion(
                fusion, drop_metadata=drop_metadata) for fusion in fusions)

        for insertion in insertions:
            yield insertion



# -*- coding: utf-8 -*-
"""Implements indexer for building Tophat references."""

# pylint: disable=wildcard-import,redefined-builtin,unused-wildcard-import
from __future__ import absolute_import, division, print_function
from builtins import *
# pylint: enable=wildcard-import,redefined-builtin,unused-wildcard-import

import tempfile
import shutil

import pathlib2 as pathlib

from imfusion.external.bowtie import bowtie_index
from imfusion.external.tophat import tophat2_index_transcriptome

from .base import Indexer, register_indexer, Reference


class TophatIndexer(Indexer):
    """Indexer that builds references for the Tophat-Fusion aligner.

    Performs the same steps as the base ``Indexer`` class, but additionally
    generates an index for alignment with Tophat-Fusion using Bowtie and
    Tophat2.
    """

    def __init__(self, logger=None, skip_index=False):
        super().__init__(logger=logger, skip_index=False)

    @property
    def _reference_class(self):
        """Reference class to use for this indexer."""
        return TophatReference

    @property
    def dependencies(self):
        """External dependencies required by this indexer."""
        return ['bowtie-build', 'tophat2']

    def _build_indices(self, reference):
        # type: (TophatReference) -> None

        # Build bowtie index.
        self._logger.info('Building bowtie index')

        bowtie_idx_path = reference.index_path
        bowtie_log_path = reference.base_path / 'bowtie.log'
        bowtie_index(
            reference_path=reference.fasta_path,
            output_base_path=bowtie_idx_path,
            log_path=bowtie_log_path)

        # Build transcriptome index.
        self._logger.info('Building transcriptome index')
        transcriptome_path = reference.transcriptome_path
        transcriptome_log_path = reference.base_path / 'transcriptome.log'
        tophat2_index_transcriptome(
            bowtie_index_path=bowtie_idx_path,
            gtf_path=reference.gtf_path,
            output_base_path=transcriptome_path,
            log_path=transcriptome_log_path)


register_indexer('tophat', TophatIndexer)


class TophatReference(Reference):
    """Tophat Reference class.

    Defines paths to files within the Tophat-Fusion reference. Compared to the
    base reference, this class adds an additional path to the transcriptome
    index that is used by Tophat during alignment.
    """

    @property
    def transcriptome_path(self):
        # type: (...) -> pathlib.Path
        """Path to transcriptome index."""
        return self._reference / 'transcriptome'

    @property
    def index_path(self):
        # type: (...) -> pathlib.Path
        """Path to index."""
        return self._reference / 'reference'



# -*- coding: utf-8 -*-
"""Tests for imfusion.insertions.aligners.tophat module."""

# pylint: disable=wildcard-import,redefined-builtin,unused-wildcard-import
from __future__ import absolute_import, division, print_function
from builtins import *
# pylint: enable=wildcard-import,redefined-builtin,unused-wildcard-import

import argparse
import shutil

from pathlib2 import Path

from future.utils import native_str
import pytest

from imfusion.insertions.aligners import tophat
from imfusion.model import Insertion
from imfusion.util.frozendict import frozendict

# pylint: disable=no-self-use,redefined-outer-name


@pytest.fixture
def read_paths():
    """Example read paths."""
    return (Path('a.fastq.gz'), Path('b.fastq.gz'))


@pytest.fixture
def tophat_path():
    """Return path to fusions.out file."""
    return pytest.helpers.data_path('fusions.out', relative_to=__file__)


@pytest.fixture
def tophat_reference():
    """Returns example reference for Tophat2."""
    ref_path = pytest.helpers.data_path('reference', relative_to=__file__)
    return tophat.TophatReference(ref_path)


@pytest.fixture
def tophat_output_dir(tmpdir, tophat_path):
    """Simulates Tophat2 output directory."""

    # Create directories.
    output_dir = Path(native_str(tmpdir / 'out'))

    tophat_dir = output_dir / '_tophat'
    tophat_dir.mkdir(parents=True)

    # Copy / simulate aligner output files.
    shutil.copy(native_str(tophat_path),
                native_str(tophat_dir / 'fusions.out'))  # yapf: disable

    pytest.helpers.touch(tophat_dir / 'Aligned.sortedByCoord.out.bam')

    return output_dir


@pytest.fixture
def cmdline_args(tmpdir):
    """Example command line arguments."""
    return [
        '--fastq', 'a.fastq.gz', '--fastq2', 'b.fastq.gz', '--reference',
        native_str(tmpdir), '--output_dir', '/path/to/out'
    ]


class TestTophatAligner(object):
    """Tests for the TophatAligner class."""

    def test_dependencies(self, tophat_reference):
        """Test dependencies for various configurations."""

        # Basic call.
        aligner = tophat.TophatAligner(tophat_reference)
        assert set(aligner.dependencies) == {'tophat2', 'bowtie'}

        # With assembly.
        aligner = tophat.TophatAligner(tophat_reference, assemble=True)
        assert set(aligner.dependencies) == {'tophat2', 'bowtie', 'stringtie'}

    def test_identify_insertions(self, read_paths, tophat_reference,
                                 tophat_output_dir, mocker):
        """Integration test using data from an example run."""

        # TODO: Check assembly?

        # Mock star_align call.
        tophat_mock = mocker.patch.object(tophat, 'tophat2_align')
        mocker.patch.object(tophat.util, 'count_lines', return_value=8e6)

        # Call identify insertions.
        fastq, fastq2 = read_paths

        aligner = tophat.TophatAligner(tophat_reference)
        ins = list(
            aligner.identify_insertions(
                fastq, tophat_output_dir, fastq2_path=fastq2))

        # Check call to star_align.

        tophat_mock.assert_called_once_with(
            fastq_path=fastq,
            fastq2_path=fastq2,
            output_dir=tophat_output_dir / '_tophat',
            index_path=tophat_reference.index_path,
            extra_args={
                '--fusion-search': (),
                '--transcriptome-index':
                (native_str(tophat_reference.transcriptome_path), ),
                '--num-threads': (1, ),
                '--bowtie1': (),
                '--fusion-anchor-length': (12, )
            })

        # Check result, including specific Cblb insertion.
        assert len(ins) == 7

        assert ins[2].id == 'INS_4'
        assert ins[2].seqname == '16'
        assert ins[2].position == 52141093
        assert ins[2].strand == -1
        assert ins[2].support_junction == 462
        assert ins[2].support_spanning == 103
        assert ins[2].support == 565
        assert ins[2].metadata['gene_id'] == 'ENSMUSG00000022637'
        assert ins[2].metadata['transposon_anchor'] == 1539
        assert ins[2].metadata['feature_name'] == 'En2SA'
        assert ins[2].metadata['gene_name'] == 'Cblb'
        assert ins[2].metadata['feature_type'] == 'SA'
        assert ins[2].metadata['gene_strand'] == 1
        assert ins[2].metadata['orientation'] == 'antisense'
        assert ins[2].metadata['ffpm_junction'] == 231.0
        assert ins[2].metadata['ffpm_spanning'] == 51.5
        assert ins[2].metadata['ffpm'] == 282.5

    def test_from_args_basic(self, cmdline_args):
        """Tests creation with minimal arguments."""

        # Setup parser.
        parser = argparse.ArgumentParser()
        tophat.TophatAligner.configure_args(parser)

        # Construct aligner.
        args = parser.parse_args(cmdline_args)
        aligner = tophat.TophatAligner.from_args(args)

        # Check args.
        assert args.fastq == Path('a.fastq.gz')
        assert args.fastq2 == Path('b.fastq.gz')
        assert args.output_dir == Path('/path/to/out')

        # Check aligner.
        # pylint: disable=w0212
        assert not aligner._assemble
        assert aligner._assemble_args == {}
        assert aligner._min_flank == 12
        assert aligner._threads == 1
        assert aligner._extra_args == {}
        assert aligner._filter_features
        assert aligner._filter_orientation
        assert aligner._filter_blacklist is None

    def test_from_args_extra_args(self, cmdline_args):
        """Tests creation with extra options."""

        cmdline_args += [
            '--tophat_threads', '5', '--tophat_min_flank', '20',
            '--tophat_args', "--limitBAMsortRAM 2000", '--assemble',
            '--no_filter_orientation', '--no_filter_feature',
            '--blacklisted_genes', 'En2'
        ] # yapf:disable

        # Setup parser.
        parser = argparse.ArgumentParser()
        tophat.TophatAligner.configure_args(parser)

        # Construct aligner.
        args = parser.parse_args(cmdline_args)
        aligner = tophat.TophatAligner.from_args(args)

        # Check args.
        assert args.fastq == Path('a.fastq.gz')
        assert args.fastq2 == Path('b.fastq.gz')
        assert args.output_dir == Path('/path/to/out')

        # Check aligner.
        # pylint: disable=w0212
        assert aligner._assemble
        assert aligner._assemble_args == {}
        assert aligner._min_flank == 20
        assert aligner._threads == 5
        assert aligner._extra_args == {'--limitBAMsortRAM': ('2000', )}
        assert not aligner._filter_features
        assert not aligner._filter_orientation
        assert aligner._filter_blacklist == ['En2']



# -*- coding: utf-8 -*-
"""Script for merging insertion and expression datasets."""

# pylint: disable=wildcard-import,redefined-builtin,unused-wildcard-import
from __future__ import absolute_import, division, print_function
from builtins import *
# pylint: enable=wildcard-import,redefined-builtin,unused-wildcard-import

import argparse

from pathlib2 import Path

import imfusion
from imfusion.merge import merge_samples
from imfusion.model import Insertion


def main():
    """Main function of im-fusion-merge."""

    args = _parse_args()

    # Use directory names to name samples if no names given.
    names = args.names or [fp.name for fp in args.sample_dirs]

    # Merge samples into single dataset.
    merged_ins, merged_expr = merge_samples(
        args.sample_dirs,
        sample_names=names,
        with_expression=args.output_expression is not None)

    # Write output(s).
    Insertion.to_csv(str(args.output), merged_ins, sep='\t', index=False)

    if args.output_expression is not None:
        merged_expr.to_csv(str(args.output_expression), sep='\t', index=True)


def _parse_args():
    """Parses command-line arguments for im-fusion-merge."""

    parser = argparse.ArgumentParser()

    parser.add_argument(
        '--version',
        action='version',
        version='IM-Fusion ' + imfusion.__version__)

    parser.add_argument(
        '--sample_dirs',
        type=Path,
        nargs='+',
        required=True,
        help='Path to sample directories.')

    parser.add_argument(
        '--output',
        type=Path,
        required=True,
        help='Output path for merged insertion file.')

    parser.add_argument(
        '--names',
        nargs='+',
        required=False,
        default=None,
        help='Alternative sample names to use for samples in merged dataset.')

    parser.add_argument(
        '--output_expression',
        type=Path,
        default=None,
        help='Output path for merged expression file.')

    return parser.parse_args()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""Tests for imfusion.build.indexers.star module."""

# pylint: disable=wildcard-import,redefined-builtin,unused-wildcard-import
from __future__ import absolute_import, division, print_function
from builtins import *
# pylint: enable=wildcard-import,redefined-builtin,unused-wildcard-import

import argparse

import pyfaidx
import pytest

from pathlib2 import Path

from imfusion.build.indexers import star

# pylint: disable=no-self-use,redefined-outer-name


class TestStarIndexer(object):
    """Tests for StarIndexer class."""

    def test_dependencies(self):
        """Check dependencies."""

        indexer = star.StarIndexer()
        assert indexer.dependencies == ['STAR']

    def test_build(self, build_kws, mocker):
        """Tests build using example files."""

        # Mock STAR call.
        mock = mocker.patch.object(star, 'star_index')

        # Build reference.
        indexer = star.StarIndexer()
        indexer.build(**build_kws)

        # Check if reference files exist.
        ref = star.StarReference(build_kws['output_dir'])

        assert ref.base_path.exists()
        assert ref.fasta_path.exists()
        assert ref.gtf_path.exists()
        assert ref.indexed_gtf_path.exists()
        # assert ref.index_path.exists()
        assert ref.transposon_name == 'T2onc'
        assert ref.transposon_path.exists()
        assert ref.features_path.exists()

        # Check presence of augmented reference sequences.
        refseq = pyfaidx.Fasta(str(ref.fasta_path))
        assert sorted(refseq.keys()) == ['1', '2', 'T2onc']

        # Check call to STAR for building the index.
        mock.assert_called_once_with(
            fasta_path=ref.fasta_path,
            gtf_path=ref.gtf_path,
            output_dir=ref.index_path,
            log_path=build_kws['output_dir'] / 'star.log',
            overhang=100,
            threads=1)

    def test_from_args(self, cmdline_args):
        """Tests creation from command line."""

        # Setup parser.
        parser = argparse.ArgumentParser()
        star.StarIndexer.configure_args(parser)

        # Construct aligner.
        args = parser.parse_args(cmdline_args)
        indexer = star.StarIndexer.from_args(args)

        # Check args.
        assert args.reference_seq == Path('/path/to/ref')
        assert args.reference_gtf == Path('/path/to/gtf')
        assert args.transposon_seq == Path('/path/to/tr')
        assert args.transposon_features == Path('/path/to/feat')

        # Check instance.
        assert indexer._overhang == 100
        assert indexer._threads == 1

    def test_from_args_extra(self, cmdline_args):
        """Tests creation from command line using extra args."""

        cmdline_args += ['--star_overhang', '50', '--star_threads', '10']

        # Setup parser.
        parser = argparse.ArgumentParser()
        star.StarIndexer.configure_args(parser)

        # Construct aligner.
        args = parser.parse_args(cmdline_args)
        indexer = star.StarIndexer.from_args(args)

        # Check instance.
        assert indexer._overhang == 50
        assert indexer._threads == 10


class TestStarReference(object):
    """Tests for StarReference class."""

    def test_non_existing(self, tmpdir):
        """Tests non-existing reference"""

        with pytest.raises(ValueError):
            star.StarReference(tmpdir / 'test')



# -*- coding: utf-8 -*-
"""Implements indexer for building STAR references."""

# pylint: disable=wildcard-import,redefined-builtin,unused-wildcard-import
from __future__ import absolute_import, division, print_function
from builtins import *
# pylint: enable=wildcard-import,redefined-builtin,unused-wildcard-import

from typing import Any

import toolz

from imfusion.external.star import star_index
from .base import Indexer, Reference, register_indexer


class StarIndexer(Indexer):
    """Indexer that builds references for the STAR aligner.

    Performs the same steps as the base ``Indexer`` class, but additionally
    generates an index for alignment with STAR using STAR's ``genomeGenerate``
    command. Special attention should be paid to the ``overhang`` parameter,
    which defines the overhang used by STAR in the build reference
    (see the ``sjdbOverhang`` parameter in the STAR documentation for more
    details). Ideally, the value for this parameter should be one less than the
    length of the used reads.
    """

    def __init__(self, logger=None, skip_index=False, overhang=100, threads=1):
        # type: (Any, int) -> None
        super().__init__(logger=logger, skip_index=skip_index)
        self._overhang = overhang
        self._threads = threads

    @property
    def _reference_class(self):
        """Reference class to use for this indexer."""
        return StarReference

    @property
    def dependencies(self):
        """External dependencies required by this indexer."""
        return ['STAR']

    def _build_indices(self, reference):
        # type: (StarReference) -> None

        self._logger.info('Building STAR index')

        star_index(
            fasta_path=reference.fasta_path,
            gtf_path=reference.gtf_path,
            output_dir=reference.index_path,
            overhang=self._overhang,
            threads=self._threads,
            log_path=reference.base_path / 'star.log')

    @classmethod
    def configure_args(cls, parser):
        """Configures an argument parser for the Indexer.

        Used by ``imfusion-build`` to configure the sub-command for
        this indexer (if registered as an Indexer using the
        ``register_indexer`` function).

        Parameters
        ----------
        parser : argparse.ArgumentParser
            Argument parser to configure.

        """
        super().configure_args(parser)
        star_group = parser.add_argument_group('STAR arguments')
        star_group.add_argument('--star_overhang', type=int, default=100)
        star_group.add_argument('--star_threads', type=int, default=1)

    @classmethod
    def _parse_args(cls, args):
        super_args = super()._parse_args(args)
        return toolz.merge(super_args, {
            'overhang': args.star_overhang,
            'threads': args.star_threads
        })


class StarReference(Reference):
    """Star Reference class.

    Defines paths to files within the STAR reference. Currently the same
    as the base ``Reference`` class.
    """
    pass


register_indexer('star', StarIndexer)



"""
Frozendict implementation (version 1.2), obtained from
slezica/python-frozendict on Github
(https://github.com/slezica/python-frozendict).

Copyright (c) 2012 Santiago Lezica

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import collections
import operator
import functools
import sys

try:
    from collections import OrderedDict
except ImportError:  # python < 2.7
    OrderedDict = NotImplemented

iteritems = getattr(dict, 'iteritems', dict.items)  # py2-3 compatibility


class frozendict(collections.Mapping):
    """
    An immutable wrapper around dictionaries that implements the complete
    :py:class:`collections.Mapping` interface. It can be used as a drop-in
    replacement for dictionaries where immutability is desired.
    """

    dict_cls = dict

    def __init__(self, *args, **kwargs):
        self._dict = self.dict_cls(*args, **kwargs)
        self._hash = None

    def __getitem__(self, key):
        return self._dict[key]

    def __contains__(self, key):
        return key in self._dict

    def copy(self, **add_or_replace):
        return self.__class__(self, **add_or_replace)

    def __iter__(self):
        return iter(self._dict)

    def __len__(self):
        return len(self._dict)

    def __repr__(self):
        return '<%s %r>' % (self.__class__.__name__, self._dict)

    def __hash__(self):
        if self._hash is None:
            h = 0
            for key, value in iteritems(self._dict):
                h ^= hash((key, value))
            self._hash = h
        return self._hash


class FrozenOrderedDict(frozendict):
    """
    A frozendict subclass that maintains key order
    """

    dict_cls = OrderedDict


if OrderedDict is NotImplemented:
    del FrozenOrderedDict



"""Module containing functions for calling stringtie."""

from pathlib2 import Path

from .util import run_command, flatten_arguments


def stringtie_assemble(
        bam_path,  # type: pathlib2.Path
        gtf_path,  # type: pathlib2.Path
        output_path,  # type: pathlib2.Path
        extra_args=None,  # type: Dict[str, Iterable[Any]]
        log_path=None  # type: pathlib2.Path
):  # type: (...) -> None
    """Runs stringtie to assemble transripts."""

    extra_args = extra_args or {}

    # Inject args.
    extra_args['-G'] = (str(gtf_path), )
    extra_args['-o'] = (str(output_path), )

    # Format args.
    extra_cmdline_args = flatten_arguments(extra_args)
    extra_cmdline_args = [str(arg) for arg in extra_cmdline_args]

    # Assemble full argument list.
    args = ['stringtie', str(bam_path)] + extra_cmdline_args
    run_command(args, log_path=log_path)



# -*- coding: utf-8 -*-
"""Contains utility functions for building references."""

# pylint: disable=wildcard-import,redefined-builtin,unused-wildcard-import
from __future__ import absolute_import, division, print_function
from builtins import *
# pylint: enable=wildcard-import,redefined-builtin,unused-wildcard-import

import re
from typing import Iterable, Tuple

import pathlib2 as pathlib

import pandas as pd
import pyfaidx

from imfusion.util.tabix import GtfIterator

REGION_REGEX = re.compile(r'(?P<chromosome>\w+):(?P<start>\d+)-(?P<end>\d+)')


def build_reference(
        reference_path,  # type: pathlib.Path
        transposon_path,  # type: pathlib.Path
        output_path,  # type: pathlib.Path
        blacklisted_regions=None  # type: Iterable[Tuple[str, int, int]]
):  # type: (...) -> None
    """Builds an augmented reference genome containing the transposon.

    Optionally, specific regions of the original reference can be blacklisted if
    these regions contain sequences that are also in the transposon sequence,
    as these may be problematic in the alignment. For example, in our
    Sleeping Beauty screen we masked regions in the Foxf2 and in En2 genes,
    as the T2onc2 transposon we used contains sequences from these genes.
    Sequences in blacklisted regions are replaced by Ns, effectively removing
    the regions from the reference.

    Parameters
    ----------
    reference_path : pathlib.Path
        Path to the fasta file containing the reference genome.
    transposon_path : pathlib.Path
        Path to the fasta file containg the transposon sequence.
    output_path : pathlib.Path
        Output path for the augmented reference genome (which is written
        as a new fasta file).
    blacklisted_regions : List[(chrom, start, ends)]
        List of regions that should be blacklisted in the augmented reference
        genome. Regions are specified as a tuple of (chr, start, end).
    """

    # Combine reference and transposon sequences.
    _concatenate_fastas(
        [reference_path, transposon_path], output_path=output_path)

    # Mask any blacklisted regions.
    if len(blacklisted_regions) > 0:
        mask_reference(output_path, blacklisted_regions)


def _concatenate_fastas(fasta_paths, output_path):
    # type: (Iterable[pathlib.Path], pathlib.Path) -> None
    """Concatenates multiple fasta files into a single file.

    This function combines multiple fasta files into a single output
    file. It is mainly used to generate the combined reference genome
    that contains both the reference genome sequence and the
    transposon sequence.

    Parameters
    ----------
    fasta_paths : Iterable[pathlib.Path]
        Paths to fasta files that should be concatentated.
    output_path : pathlib.Paths
        Path for the combined output file.

    """

    with output_path.open('wt') as fileobj:
        for fasta_path in fasta_paths:
            # Open fasta file.
            fasta = _open_fasta_file(fasta_path)

            # Read/write sequences.
            for record in fasta:
                _write_fasta_record(
                    fileobj, name=record.long_name, sequence=str(record))


def _open_fasta_file(file_path, **kwargs):
    """Opens a fasta file using pyfaidx."""

    try:
        fasta = pyfaidx.Fasta(str(file_path), **kwargs)
    except pyfaidx.FastaIndexingError:
        raise ValueError('Failed to read fasta file ({}). Ensure that '
                         'the file is a valid fasta file.'.format(
                             str(file_path)))
    return fasta


def _write_fasta_record(fileobj, name, sequence, max_width=80):
    """Writes a fastq record to an opened file."""

    # Group sequence into chunks of max_width.
    chunked = '\n'.join((sequence[i:i + max_width]
                         for i in range(0, len(sequence), max_width)))

    # Write record to file.
    fileobj.write(u'>{}\n{}\n'.format(name, chunked))


def mask_reference(reference_path, blacklist_regions):
    # type: (pathlib.Path, Iterable[Tuple[str, int, int]]) -> None
    """Masks blacklisted regions in a given reference sequence.

    This function removes blacklisted regions from a given reference genome.
    Blacklisted regions are removed by replacing their original sequence
    with a sequences of 'N' nucleotides.

    Parameters
    ----------
    refseq_path : pathlib.Path
        Path to the reference sequence in fasta format.
    blacklist_regions : Iterable[tuple(str, int, int)]:
        List of regions that should be blacklisted in the reference
        sequence. Regions should be specified as a tuple of (chromosome,
        start_position, end_position). For example: ('1', 2000, 2200).

    """

    # Open the reference as a mutable sequence.
    ref = _open_fasta_file(reference_path, mutable=True)

    # Mask regions.
    for seqname, start, end in blacklist_regions:
        ref[seqname][start:end] = 'N' * (end - start)

    # Close file.
    ref.close()


def regions_from_strings(region_strs):
    # type: (Iterable[str]) -> List[Tuple[str, int, int]]
    """Builds blacklist region list for region strings.

    Parses region strings into a list of blacklist region tuples.
    Region strings should be provided in the following format:
    'chromosome:start-end'. For example, 'X:1000-2000' denotes a
    a region on chromosome X from position 1000 to 2000.

    Parameters
    ----------
    region_strs : Iterable[str]
        List of region strings.

    Returns
    -------
    Listp[Tuple(str, int, int]]
        List of blacklist region tuples.

    """

    return [_parse_region_str(s) for s in region_strs]


def _parse_region_str(region_str):
    # type: (str) -> Tuple[str, int, int]

    match = REGION_REGEX.search(region_str)

    if match is None:
        raise ValueError('Unable to parse region {!r}. Ensure that region '
                         'specifications are in the required format '
                         '(chromosome:start-end).'.format(region_str))
    else:
        groups = match.groupdict()
        return groups['chromosome'], int(groups['start']), int(groups['end'])


def regions_from_genes(gene_ids, gtf_path):
    # type: (List[str], pathlib.Path) -> List[Tuple[str, int, int]]
    """Builds blacklist frame for given genes.

    Returns a list of blacklist regions encompassing the regions
    spanned by the genes corresponding to the given gene ids.

    Parameters
    ----------
    gene_ids : List[str]
        List of (Ensembl) gene ids.
    gtf_path : pathlib.Path
        Path to gtf file.

    Returns
    -------
    Tuple[str, int, int]
        List of blacklist region tuples.

    """

    # Fetch genes.
    gtf_iter = GtfIterator(gtf_path)
    gene_records = list(gtf_iter.fetch_genes(gene_ids))

    # Check for missing genes.
    for gene_id, record in zip(gene_ids, gene_records):
        if record is None:
            raise ValueError(
                'Gene {!r} not found in reference gtf. Make sure that the '
                'correct gene ID is being used (corresponding with the '
                'gene_id annotation in the GTF file).'.format(gene_id))

    return [(gene.contig, gene.start, gene.end) for gene in gene_records]


def check_feature_file(feature_path):
    """Checks if a transposon feature frame is valid."""

    transposon_features = pd.read_csv(str(feature_path), sep='\t')

    # Check dataframe columns.
    req_cols = {'name', 'start', 'end', 'strand', 'type'}
    missing_cols = req_cols - set(transposon_features.columns)

    if len(missing_cols) > 0:
        raise ValueError('Missing required columns {!r} in the transposon '
                         'file. Make sure that the file is a properly '
                         'formatted tab-separated file (see documentation for '
                         'more details).'.format(missing_cols))

    # Check if we have any features.
    if len(transposon_features) < 1:
        raise ValueError('Transposon feature file is empty.')

    # Check if we have a valid feature.
    num_valid = transposon_features['type'].isin({'SA', 'SD'}).sum()
    if num_valid < 1:
        raise ValueError('No valid features (with type = SD or SA)'
                         'were found in the transposon feature file.')



"""Module containing functions for calling featureCounts."""

from .util import run_command, flatten_arguments


def feature_counts(
        bam_files,  # type: Iterable[pathlib.Path]
        gtf_path,  # type: pathlib.Path
        output_path,  # type: pathlib.Path
        extra_kws=None,  # type: Dict[str, Any]
        log_path=None  # type: pathlib.Path
):  # type: (...) -> None
    """Generates counts using featureCounts.

    Parameters
    ----------
    bam_files : list[pathlib.Path]
        List of paths to the bam files for which counts should be generated.
    gtf_path : pathlib.Path
        Path to the GTF file containing gene features.
    output_path : pathlib.Path
        Path to output file.
    extra_kws : dict[str, tuple]:
        Dictionary of extra arguments that should be passed to feature counts.
        Keys should correspond to argument names (including dashes),
        values should be tuples containing the argument values.

    Returns
    -------
    pandas.Dataframe
        DataFrame containing feature counts for the given bam files. The rows
        correspond to the counted features, the columns correspond to the
        index values (chomosome, position etc.) and the bam files.

    """

    extra_kws = extra_kws or {}

    # Run feature counts.
    args = (['featureCounts'] + flatten_arguments(extra_kws) +
            ['-a', str(gtf_path), '-o', str(output_path)] +
            [str(bf) for bf in bam_files])

    run_command(args=args, log_path=log_path)



"""Module containing functions for calling sambamba."""

from .util import run_command


def sambamba_sort(input_bam,
                  output_bam,
                  threads=1,
                  tmp_dir=None,
                  log_path=None):
    """Sorts bam file using sambamba.

    Parameters
    ----------
    input_bam : Path
        Path to input (unsorted) bam file.
    output_bam : Path
        Path for output (sorted) bam file.
    threads : int
        Number of threads to use.
    tmp_dir : Path
        Temporary directory to use.
    log_path : Path
        Path to log file.

    """

    if tmp_dir is None:
        tmp_dir = output_bam.parent / '_tmp'

    args = [
        'sambamba', 'sort', '-o', str(output_bam), '--tmpdir=' + str(tmp_dir),
        '-t', str(threads), str(input_bam)
    ]

    run_command(args=args, log_path=log_path)



"""Script for building StarFusion indices using FusionFilter."""

import argparse
import gzip
import logging
from pathlib import Path
import shutil
import subprocess

FORMAT = "[%(asctime)-15s] %(message)s"
logging.basicConfig(format=FORMAT, level=logging.INFO)


def main():
    """Main function."""

    args = parse_args()

    # Create output and temp dirs.
    tmp_dir = args.output_dir / '_tmp'
    tmp_dir.mkdir(parents=True, exist_ok=False)

    # Filter the patch chromosomes from the gtf, as these are
    # likely not present in the fasta.
    logging.info('- Generating cDNA sequences')

    gtf_path = tmp_dir / 'ref.gtf'
    with gtf_path.open('wb') as file_:
        check_call(['grep', '-v', r'^\(MG\|JH\|GL\)', str(args.gtf)],
                   stdout=file_)

    # Create cDNA_seqs file.
    cdna_path = tmp_dir / 'cDNA_seqs.fa'
    with cdna_path.open('wb') as file_:
        script_path = args.ff_path / 'util' / 'gtf_file_to_cDNA_seqs.pl'
        check_call(['perl', str(script_path), str(gtf_path), str(args.fasta)],
                   stdout=file_)

    # Build masked cDNA_seqs file using RepeatMasker.
    # Note: requires library to be installed from http://www.girinst.org.
    logging.info('- Masking repeats')

    masked_path = cdna_path.with_suffix('.fa.masked')
    check_call([str(args.rm_path / 'RepeatMasker'), '-pa', str(args.threads), '-s',
                '-species', 'mouse', '-xsmall', str(cdna_path)])

    # Create blastpairs.
    logging.info('- Creating blast pairs')

    check_call(['makeblastdb', '-in', str(masked_path), '-dbtype', 'nucl'])

    pair_path = tmp_dir / 'blast_pairs.outfmt6'
    with pair_path.open('wb') as file_:
        check_call(['blastn',
                    '-query', str(cdna_path),
                    '-db', str(masked_path),
                    '-max_target_seqs', '10000',
                    '-outfmt', '6',
                    '-evalue', '1e-3',
                    '-lcase_masking',
                    '-num_threads', str(args.threads),
                    '-word_size', '11'],
                   stdout=file_)

    pair_gz_path = pair_path.with_suffix('.gene_syms.outfmt6.gz')
    with gzip.open(str(pair_gz_path), 'wb') as file_:
        script_path = (args.ff_path / 'util' /
                       'blast_outfmt6_replace_trans_id_w_gene_symbol.pl')
        check_call(['perl', str(script_path), str(cdna_path), str(pair_path)],
                   stdout=file_)

    # Prepare library.
    logging.info('- Preparing library')
    script_path = args.ff_path / 'util' / 'prep_genome_lib.pl'
    check_call(['perl', str(script_path),
                '--genome_fa', str(args.fasta),
                '--gtf', str(gtf_path),
                '--blast_pairs', str(pair_gz_path),
                '--cdna_fa', str(cdna_path),
                '--CPU', str(args.threads),
                '--max_readlength', str(args.read_length),
                '--output_dir', str(args.output_dir)])

    # shutil.rmtree(str(args.tmp_dir))

def parse_args():
    """Parses command line arguments."""

    parser = argparse.ArgumentParser()

    parser.add_argument('--fasta', required=True, type=Path)
    parser.add_argument('--gtf', required=True, type=Path)
    parser.add_argument('--output_dir', required=True, type=Path)
    
    parser.add_argument('--ff_path', required=False, default='', type=Path)
    parser.add_argument('--rm_path', required=False, default='', type=Path)

    parser.add_argument('--threads', type=int, default=1)
    parser.add_argument('--read_length', type=int, default=100)

    return parser.parse_args()


def check_call(args, verbose=True, **kwargs):
    """Wrapper function for check_call."""

    if verbose:
        logging.info('Running command - "%s"', ' '.join(args))

    subprocess.check_call(args, **kwargs)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""Implements base indexer classes, used for building references."""

# pylint: disable=wildcard-import,redefined-builtin,unused-wildcard-import
from __future__ import absolute_import, division, print_function
from builtins import *
# pylint: enable=wildcard-import,redefined-builtin,unused-wildcard-import

# pylint: disable=W0611

import argparse
import logging
import shutil
from typing import Any, Type, Tuple

import pathlib2 as pathlib

import pyfaidx

from imfusion.external.util import check_dependencies
from imfusion.util import tabix

from .. import util as build_util

_indexer_registry = {}


def register_indexer(name, indexer):
    # type: (str, Type[Indexer]) -> None
    """Registers indexer under given name."""
    _indexer_registry[name] = indexer


def get_indexers():
    # type: (...) -> Dict[str, Type[Indexer]]
    """Returns dict of available indexers."""
    return dict(_indexer_registry)


class Indexer(object):
    """Base Indexer class.

    This base class defines the interface for Indexer classes and provides
    basic functionality for building augmented reference genomes. Functionality
    for building the indices for alignment is provided by sub-classes
    (as this is aligner specific), typically by overriding the protected
    ``_build_indices`` method.

    The main interface of the class is formed by the ``check_dependencies``
    and ``build`` methods. The former is used to check whether required
    external dependencies are present in ``$PATH``, whilst the latter is
    called with the required reference files to build the actual reference.
    On completion, the ``build`` method returns a ``Reference`` object, which
    defines paths to the various reference files. The class of this object is
    defined by the ``_reference_class`` method, which may be overridden to
    return aligner-specific ``Reference`` subclasses.

    Additionally, the class provides the methods ``configure_args`` and
    ``from_args``, which are used to configure the argument parser for
    ``imfusion-build`` and to instantate instances from parsed arguments.
    Subclasses with extra options should override the protected ``_parse_args``
    method, which extracts constructor parameters from argparse arguments
    for ``from_args``, allowing for flexible interpretation of these arguments.

    Parameters
    ----------
    logger : logging.Logger
        Logger to be used for logging messages.
    skip_index : bool
        Whether to skip building the index. Mostly used for debugging purposes,
        as this typically results in an incomplete reference.

    """

    def __init__(self, logger=None, skip_index=False):
        # type: (Any) -> None

        self._logger = logger or logging.getLogger()
        self._skip_index = skip_index

    @property
    def _reference_class(self):  # type: (...) -> Type[Reference]
        """Reference class to use for this indexer."""
        return Reference

    @property
    def dependencies(self):  # type: (...) -> List[str]
        """External dependencies required by Indexer."""
        return []

    def check_dependencies(self):
        """Checks if all required external dependencies are in ``$PATH``.

        Raises a ValueError if any dependencies are missing.
        """

        check_dependencies(self.dependencies)

    def build(
            self,
            refseq_path,  # type: pathlib.Path
            gtf_path,  # type: pathlib.Path
            transposon_path,  # type: pathlib.Path
            transposon_features_path,  # type: pathlib.Path
            output_dir,  # type: pathlib.Path
            blacklist_regions=None,  # type: List[Tuple[str, int, int]]
            blacklist_genes=None  # type: List[str]
    ):  # type: (...) -> None
        """Builds an augmented reference containing the transposon sequence.

        The reference is built by first copying the original reference
        sequence into the output directory and then augmenting this reference
        with the transposon sequence. If needed, blacklisted genes/regions are
        then masked within this reference. Next, other supporting files
        (such as the reference GTF) are copied and indexed as needed. As a
        final step, aligner-specific indices are built for the augmented
        reference.

        Parameters
        ----------
        refseq_path : Path
            Path to the reference sequence (in Fasta format).
        gtf_path : Path
            Path to the reference GTF file.
        transposon_path : Path
            Path to the transposon sequence (in Fasta format).
        transposon_features_path : Path
            Path to the transposon features (tab-separated file).
        output_dir : Path
            Path to the output directory. Should not yet exist.
        blacklist_regions: List[Tuple[str, int, int]]
            List or regions that should be blacklisted, specified as a tuple
            of (chromosome, start, end).
        blacklist_genes : List[str]
            List of genes that should be blacklisted. Should correspond to
            ``gene_id`` entries in the passed GTF file.

        Returns
        -------
        Reference
            Returns a ``Reference`` object, which describes the paths to
            various files within the generated reference.

        """

        # Create output directory.
        try:
            output_dir.mkdir(parents=True, exist_ok=False)
        except FileExistsError as ex:
            ex.strerror = 'Output directory already exists'
            raise ex

        # Use dummy Reference instance for paths.
        reference = self._reference_class(output_dir)

        # Copy transposon files.
        self._logger.info('Copying transposon files')
        shutil.copy(str(transposon_path), str(reference.transposon_path))

        build_util.check_feature_file(transposon_features_path)
        shutil.copy(str(transposon_features_path),
                    str(reference.features_path)) # yapf: disable

        # Build indexed reference gtf.
        self._logger.info('Building indexed reference gtf')
        gtf_frame = tabix.read_gtf_frame(gtf_path)
        gtf_frame = tabix.sort_gtf_frame(gtf_frame)

        tabix.write_gtf_frame(gtf_frame, reference.gtf_path)
        tabix.index_gtf(
            reference.gtf_path,
            output_path=reference.indexed_gtf_path,
            sort=False)

        # Build flattened exon gtf.
        self._logger.info('Building flattened exon gtf')
        gtf_frame_flat = tabix.flatten_gtf_frame(gtf_frame)
        tabix.write_gtf_frame(
            gtf_frame_flat, file_path=reference.exon_gtf_path)

        # Build augmented reference.
        self._logger.info('Building augmented reference')

        blacklist_regions = blacklist_regions or []
        blacklist_genes = blacklist_genes or []

        blacklist = (build_util.regions_from_strings(blacklist_regions) +
                     build_util.regions_from_genes(blacklist_genes,
                                                   reference.indexed_gtf_path))

        build_util.build_reference(
            refseq_path,
            transposon_path,
            output_path=reference.fasta_path,
            blacklisted_regions=blacklist)

        # Build any required indices using files.
        if self._skip_index:
            self._logger.warning('Skipping the building of the index. '
                                 'Reference will not be usable for alignment.')
        else:
            self._build_indices(reference)

    def _build_indices(self, reference):
        raise NotImplementedError()

    @classmethod
    def configure_args(cls, parser):
        # type: (argparse.ArgumentParser) -> None
        """Configures an argument parser for the Indexer.

        Used by ``imfusion-build`` to configure the sub-command for
        this indexer (if registered as an Indexer using the
        ``register_indexer`` function).

        Parameters
        ----------
        parser : argparse.ArgumentParser
            Argument parser to configure.

        """

        # Basic arguments.
        base_group = parser.add_argument_group('Basic arguments')

        base_group.add_argument(
            '--reference_seq',
            type=pathlib.Path,
            required=True,
            help='Path to the reference sequence (in Fasta format).')

        base_group.add_argument(
            '--reference_gtf',
            type=pathlib.Path,
            required=True,
            help='Path to the reference gtf file.')

        base_group.add_argument(
            '--transposon_seq',
            type=pathlib.Path,
            required=True,
            help='Path to the transposon sequence (in Fasta format).')

        base_group.add_argument(
            '--transposon_features',
            type=pathlib.Path,
            required=True,
            help='Path to the transposon features (tsv).')

        base_group.add_argument(
            '--output_dir', type=pathlib.Path, required=True)

        # Optional blacklist arguments.
        blacklist_group = parser.add_argument_group('Blacklist arguments')

        blacklist_group.add_argument(
            '--blacklist_regions',
            nargs='+',
            default=(),
            help='Regions of the reference to blacklist. Should '
            'be specified as \'chromosome:start-end\'.')

        blacklist_group.add_argument(
            '--blacklist_genes',
            nargs='+',
            default=(),
            help='Genes to blacklist. Should correspond with '
            'the gene ids used in the reference gtf file.')

        debug_group = parser.add_argument_group('Debugging')
        debug_group.add_argument(
            '--skip_index',
            default=False,
            action='store_true',
            help=('Whether to skip the building of the genome indices. '
                  'Mainly used for debugging purposes.'))

    @classmethod
    def _parse_args(cls, args):
        # type: (argparse.Namespace) -> Dict[str, Any]
        """Parses argparse argument to a dict."""
        return {'skip_index': args.skip_index}

    @classmethod
    def from_args(cls, args):
        # type: (argparse.Namespace) -> Indexer
        """Constructs an Indexer instance from given arguments.

        Instantiates an Indexer instance from the given argparse arguments.
        Uses the ``_parse_args`` method internally, which performs the actual
        argument-to-parameter extraction. As such, the ``_parse_args`` method
        should be overridden in any subclasses with extra parameters.

        Parameters
        ----------
        args : argparse.Namespace
            Arguments to parse.
        """
        return cls(**cls._parse_args(args))


class Reference(object):
    """Base Reference class.

    Reference classes define paths to specific files within the reference. This
    includes, for example, paths to the augmented fasta file and to the r
    eference index. This base class defines paths that should be available
    for all references. Subclasses may define additional paths that are
    specific to the corresponding aligner.

    Parameters
    ----------
    reference_path : Path
        Path to the reference directory.

    """

    def __init__(self, reference_path):
        # type: (pathlib.Path) -> None
        if not reference_path.exists():
            raise ValueError('Reference path does not exist')
        self._reference = reference_path

    @property
    def base_path(self):
        # type: (...) -> pathlib.Path
        """Path to reference base directory."""
        return self._reference

    @property
    def fasta_path(self):
        # type: (...) -> pathlib.Path
        """Path to reference sequence."""
        return self._reference / 'reference.fa'

    @property
    def gtf_path(self):
        # type: (...) -> pathlib.Path
        """Path to reference gtf."""
        return self._reference / 'reference.gtf'

    @property
    def indexed_gtf_path(self):
        # type: (...) -> pathlib.Path
        """Path to reference gtf."""
        return self._reference / 'reference.gtf.gz'

    @property
    def exon_gtf_path(self):
        # type: (...) -> pathlib.Path
        """Path to exon gtf."""
        return self._reference / 'exons.gtf'

    @property
    def index_path(self):
        # type: (...) -> pathlib.Path
        """Path to index."""
        return self._reference / 'index'

    @property
    def transposon_name(self):
        # type: (...) -> str
        """Name of transposon sequence."""
        seqs = pyfaidx.Fasta(str(self.transposon_path)).keys()
        return list(seqs)[0]

    @property
    def transposon_path(self):
        # type: (...) -> pathlib.Path
        """Name of transposon sequence."""
        return self._reference / 'transposon.fa'

    @property
    def features_path(self):
        # type: (...) -> pathlib.Path
        """Path to transposon features."""
        return self._reference / 'features.txt'



# -*- coding: utf-8 -*-
"""Unit tests for functions calling bowtie externally."""

from pathlib2 import Path

import pytest

from imfusion.external import star

# pylint: disable=no-self-use,redefined-outer-name


class TestStarIndex(object):
    """Unit tests for the star_index function."""

    def test_call(self, mocker, tmpdir):
        """Tests example call."""

        mock_run = mocker.patch.object(star, 'run_command')

        output_dir = Path(str(tmpdir / 'out'))

        star.star_index(
            Path('reference.fa'),
            gtf_path=Path('reference.gtf'),
            output_dir=output_dir,
            overhang=50)

        mock_run.assert_called_once_with(
            args=[
                'STAR', '--runMode', 'genomeGenerate', '--genomeDir',
                str(output_dir), '--genomeFastaFiles', 'reference.fa',
                '--sjdbGTFfile', 'reference.gtf', '--sjdbOverhang', '50',
                '--runThreadN', '1'
            ],
            log_path=None)


@pytest.fixture
def star_align_kws(tmpdir):
    """Example kws for star_align."""

    output_dir = Path(str(tmpdir / 'out'))

    return {
        'fastq_path': Path('sample.R1.fastq.gz'),
        'index_path': Path('index'),
        'output_dir': output_dir
    }


class TestStarAlign(object):
    """Unit tests for the star_align function."""

    def test_basic_call(self, mocker, star_align_kws):
        """Tests example call."""

        mock_run = mocker.patch.object(star, 'run_command')

        star.star_align(**star_align_kws)

        mock_run.assert_called_once_with(
            args=[
                'STAR', '--genomeDir', 'index', '--outFileNamePrefix',
                str(star_align_kws['output_dir']) + '/', '--readFilesIn',
                'sample.R1.fastq.gz', '--readFilesCommand', 'gunzip', '-c'
            ],
            log_path=None)

    def test_basic_without_gzip(self, mocker, star_align_kws):
        """Tests example without gzipped input."""

        mock_run = mocker.patch.object(star, 'run_command')

        star_align_kws['fastq_path'] = Path('sample.R1.fastq')
        star.star_align(**star_align_kws)

        mock_run.assert_called_once_with(
            args=[
                'STAR', '--genomeDir', 'index', '--outFileNamePrefix',
                str(star_align_kws['output_dir']) + '/', '--readFilesIn',
                'sample.R1.fastq'
            ],
            log_path=None)

    def test_paired_end(self, mocker, star_align_kws):
        """Tests example call with paired-end data."""

        mock_run = mocker.patch.object(star, 'run_command')

        star_align_kws['fastq2_path'] = 'sample.R2.fastq.gz'
        star.star_align(**star_align_kws)

        mock_run.assert_called_once_with(
            args=[
                'STAR', '--genomeDir', 'index', '--outFileNamePrefix',
                str(star_align_kws['output_dir']) + '/', '--readFilesIn',
                'sample.R1.fastq.gz', 'sample.R2.fastq.gz',
                '--readFilesCommand', 'gunzip', '-c'
            ],
            log_path=None)

    def test_extra_arguments(self, mocker, star_align_kws):
        """Tests example call with extra arguments."""

        mock_run = mocker.patch.object(star, 'run_command')

        star_align_kws['extra_args'] = {'--outSAMtype': 'Unsorted'}
        star.star_align(**star_align_kws)

        mock_run.assert_called_once_with(
            args=[
                'STAR', '--genomeDir', 'index', '--outFileNamePrefix',
                str(star_align_kws['output_dir']) + '/', '--readFilesIn',
                'sample.R1.fastq.gz', '--outSAMtype', 'Unsorted',
                '--readFilesCommand', 'gunzip', '-c'
            ],
            log_path=None)

    def test_multiple_threads(self, mocker, star_align_kws):
        """Tests example call with extra arguments."""

        mock_run = mocker.patch.object(star, 'run_command')

        star_align_kws['threads'] = 5
        star.star_align(**star_align_kws)

        mock_run.assert_called_once_with(
            args=[
                'STAR', '--genomeDir', 'index', '--outFileNamePrefix',
                str(star_align_kws['output_dir']) + '/', '--readFilesIn',
                'sample.R1.fastq.gz', '--readFilesCommand', 'gunzip', '-c',
                '--runThreadN', '5'
            ],
            log_path=None)



# -*- coding: utf-8 -*-
"""Implements functions for testing for Commonly Targeted Genes (CTGs)."""

# pylint: disable=wildcard-import,redefined-builtin,unused-wildcard-import
from __future__ import absolute_import, division, print_function
from builtins import *
# pylint: enable=wildcard-import,redefined-builtin,unused-wildcard-import

from typing import Tuple, Optional, Iterable, Callable, Pattern
import itertools
import logging
import operator
import re

import pathlib2 as pathlib

import numpy as np
import pandas as pd
import pyfaidx
import pysam

from intervaltree import IntervalTree
from scipy.stats import poisson

from imfusion.build import Reference
from imfusion.model import Insertion
from imfusion.util.genomic import GenomicIntervalTree
from imfusion.util.tabix import GtfIterator


def test_ctgs(
        insertions,  # type: List[Insertion]
        reference,  # type: Reference
        gene_ids=None,  # type: Set[str]
        chromosomes=None,  # type: Set[str]
        pattern=None,  # type: str
        per_sample=True,  # type: bool
        window=None  #type: Tuple[int, int]
):
    """Identifies genes that are significantly enriched for insertions (CTGs).

    This function takes a DataFrame of insertions, coming from multiple samples,
    and identifies if any genes are more frequently affected by an
    insertion than would be expected by chance. These genes are called
    Commonly Targeted Genes (CTGs). CTGs are selected by comparing the
    number of insertions within the gene to the number of insertions
    that would be expected from the background insertion rate, which is
    modeled using a Poisson distribution.

    Parameters
    ----------
    insertions : List[Insertion]
        Insertions to test.
    reference : Reference
        Reference index used by the aligner to identify insertions.
    genes : List[str]
        List of genes to test (defaults to all genes with an insertion).
    chromosomes : List[str]
        List of chromosomes to include, defaults to all chromosomes
        in the given reference GTF.
    pattern : str
        Specificity pattern of the used transposon.
    per_sample : bool
        Whether to perform the per sample test (recommended), which
        effectively collapes insertions per sample/gene combination.
        This avoids issues in which insertions that are detected
        multiple times or that may have hopped inside the gene locus
        are counted multiple times.
    window : Tuple[int, int]
        Window to include around gene (in bp). Specified as (upstream_dist,
        downstream_dist). For example: (2000, 2000) specifies in a 2KB
        window around each gene.

    Returns
    -------
    pandas.DataFrame
        Results of CTG test for tested genes. Contains two columns:
        p_value and q_value. The last column (q_value)
        represents the p-value of the gene after correcting for
        multiple testing using bonferroni correction.

    """

    # Determine gene windows using GTF.
    logging.info('Generating gene windows')
    gene_windows = _build_gene_windows(
        reference.indexed_gtf_path, window=window, chromosomes=chromosomes)

    # Subset insertions to gene intervals.
    insertions = _subset_to_windows(insertions, gene_windows)

    if gene_ids is None:
        gene_ids = set(ins.metadata['gene_id'] for ins in insertions)

    # Collapse insertions per gene/sample (recommended).
    # Corrects for hopping/multiple detection issues.
    if per_sample:
        logging.info('Collapsing insertions')
        insertions = list(_collapse_per_sample(insertions))

    # Calculate total number of pattern occurrences within intervals.
    logging.info('Counting pattern occurrences')
    reference_seq = pyfaidx.Fasta(str(reference.fasta_path))

    total = count_total(
        reference_seq, pattern=pattern, intervals=gene_windows.values())

    # Calculate p-values for each gene.
    logging.info('Calculating significance for genes')
    insertion_trees = GenomicIntervalTree.from_objects_position(
        insertions, chrom_attr='seqname')

    p_values = {
        gene_id: test_region(
            insertions=insertions,
            reference_seq=reference_seq,
            region=gene_windows[gene_id],
            total=total,
            pattern=pattern,
            filters=[lambda ins, gid=gene_id: ins.metadata['gene_id'] == gid],
            insertion_trees=insertion_trees)
        for gene_id in gene_ids
    }

    # Build result frame.
    result = pd.DataFrame.from_records(
        iter(p_values.items()), columns=['gene_id', 'p_value'])

    # Calculate corrected p-value using bonferroni correction.
    result['q_value'] = (result['p_value'] * len(result)).clip_upper(1.0)

    # Sort by q-value and p-value.
    result.sort_values(by=['q_value', 'p_value'], inplace=True)

    # Annotate with gene_name if possible.
    if 'gene_name' in insertions[0].metadata:
        name_map = {
            ins.metadata['gene_id']: ins.metadata['gene_name']
            for ins in insertions
        }
        result.insert(1, 'gene_name', result['gene_id'].map(name_map))

    # Annotate with frequency.
    frequency = (Insertion.to_frame(insertions)
                 .groupby('gene_id')['sample'].nunique()
                 .reset_index(name='n_samples'))
    result = pd.merge(result, frequency, on='gene_id', how='left')

    return result


def _build_gene_windows(
        gtf_path,  # type: pathlib.Path
        window=None,  # type: Optional[Tuple[int, int]]
        chromosomes=None  # type: Set[str]
):
    gtf_iter = GtfIterator(gtf_path)

    if chromosomes is None:
        chromosomes = set(gtf_iter.contigs)

    records = itertools.chain.from_iterable(
        gtf_iter.fetch_genes(reference=chrom) for chrom in chromosomes)

    return {rec['gene_id']: _apply_gene_window(rec, window) for rec in records}


def _apply_gene_window(
        gene,  # type: pysam.ctabixproxies.GTFProxy
        window=None  # type: Tuple[int, int]
):  # type: (...) -> Tuple[str, int, int]

    if window is None:
        return gene.contig, gene.start, gene.end,
    else:
        upstream_offset, downstream_offset = window

        if gene.strand == '-':
            start = gene.start - downstream_offset
            end = gene.end + upstream_offset
        elif gene.strand == '+':
            start = gene.start - upstream_offset
            end = gene.end + downstream_offset
        else:
            raise ValueError('Unknown value for strand')

        return gene.contig, start, end


def _subset_to_windows(
        insertions,  # type: List[Insertion]
        gene_windows  # type: Dict[str, Tuple[str, int, int]]
):  # type: (...) -> List[Insertion]
    """Subsets insertions for given gene windows."""

    # Create lookup trees.
    trees = {
        chrom: IntervalTree.from_tuples((i[1:]) for i in chrom_int)
        for chrom, chrom_int in itertools.groupby(
            sorted(gene_windows.values()), operator.itemgetter(0))
    }

    # Determine which insertions overlap tree intervals and
    # correspond to genes with known gene window.
    def _in_windows(ins, trees):
        try:
            return trees[ins.seqname].overlaps(ins.position)
        except KeyError:
            return False

    return [
        ins for ins in insertions
        if ins.metadata['gene_id'] in gene_windows and _in_windows(ins, trees)
    ]


def _collapse_per_sample(insertions):
    # Type: (List[Insertion]) -> Generator
    def _keyfunc(insertion):
        return (insertion.metadata['sample'],
                str(insertion.metadata['gene_id']))

    grouped = itertools.groupby(sorted(insertions, key=_keyfunc), key=_keyfunc)

    for _, grp in grouped:
        grp = list(grp)

        if len(grp) > 1:
            mean_pos = int(np.mean([ins.position for ins in grp]))
            yield grp[0]._replace(position=mean_pos)
        else:
            yield grp[0]


def test_region(
        insertions,  # type: List[Insertion]
        reference_seq,  # type: pyfaidx.Fasta
        region,  # type: Tuple[str, int, int]
        pattern=None,  # type: Optional[str]
        intervals=None,  # type: Optional[Iterable[Tuple[str, int, int]]]
        total=None,  # type: Optional[int]
        filters=None,  # type: Optional[List[Callable]]
        insertion_trees=None  # type: GenomicIntervalTree
):  # type: (...) -> float
    """Tests a given genomic region for enrichment in insertions."""

    if total is None:
        total = count_total(
            reference_seq, pattern=pattern, intervals=intervals)

    # Count pattern in region.
    region_count = count_region(reference_seq, region=region, pattern=pattern)

    # Sub-select insertions for region.
    if insertion_trees is None:
        insertion_trees = GenomicIntervalTree.from_objects_position(
            insertions, chrom_attr='seqname')

    region_ins = set(interval[2]
                     for interval in insertion_trees.search(*region))

    # Apply additional filter functions to insertions if given
    # (such as filtering on gene name/id for example).
    if filters is not None:
        for filter_func in filters:
            region_ins = set(ins for ins in region_ins if filter_func(ins))

    # Calculate p-value.
    x = len(list(region_ins))
    mu = len(insertions) * (region_count / total)

    # Note here we use loc=1, because we are interested in
    # calculating P(X >= x), not P(X > x) (the default
    # surivival function).
    p_val = poisson.sf(x, mu=mu, loc=1)  # type: float

    return p_val


def count_region(
        reference_seq,  # type: pyfaidx.Fasta
        region,  # type: Tuple[str, int, int]
        pattern=None  # type: Optional[str]
):  # type: (...) -> int
    """Counts occurrences of pattern within given genomic region.

    Parameters
    ----------
    reference : pyfaidx.Fasta
        Reference to count occurrences for.
    region: Tuple[str, int, int]
        Genomic region to search in.
    pattern : str
        Nucleotide sequence to count occurences for. If None, the
        length of the region is used.

    Returns
    -------
    int
        Number of occurrences of the pattern within the given region.


    """

    chrom, start, end = region
    seq = reference_seq[chrom][int(start):int(end)]

    return _count_sequence(seq, regex=_build_regex(pattern))


def _build_regex(pattern):
    # type: (str) -> Pattern[str]
    if pattern is not None:
        return re.compile(pattern + '|' + pattern[::-1])
    return None


def _count_sequence(sequence, regex=None):
    # type: (pyfaidx.Sequence, Pattern[str]) -> int
    """Counts occurrences of pattern in sequence.

    Parameters
    ----------
    sequence : pyfaidx.Sequence
        Sequence to search.
    regex : Pattern[str]
        Pattern to count.

    Returns
    -------
        int: Number of occurrences of pattern.

    """

    if regex is None:
        count = len(sequence)
    else:
        count = sum((1 for _ in regex.finditer(str(sequence))))

    return count


def count_total(
        reference_seq,  # type: pyfaidx.Sequence
        pattern=None,  # type: str
        intervals=None  # type: Iterable[Tuple[str, int, int]]
):  # type: (...) -> int
    """Counts total occurrences of pattern in reference.

    Parameters
    ----------
    reference : pyfaidx.Fasta
        Reference to count occurrences for.
    pattern : str
        Nucleotide sequence to search for. If None, the length of
        sequences is counted instead of pattern of occurrences.
    intervals : List[tuple(str, int, int)]
        List of genomic intervals to which search should be restricted.
        If None, the entire reference is used.

    Returns
    -------
    int
        Number of occurrences of the pattern within the given reference,
        or within the given intervals (if applicable).


    """

    regex = _build_regex(pattern)

    if intervals is None:
        # Simply count for the entire sequence.
        count = sum(_count_sequence(reference_seq[seq], regex=regex)
                    for seq in reference_seq.keys()) # yapf: disable
    else:
        # Flatten intervals, and then only count for sequences
        # within the flattened intervals.
        merged_intervals = list(merge_genomic_intervals(intervals))

        seqs = [
            reference_seq[chrom][start:end]
            for chrom, start, end in merged_intervals
        ]

        count = sum(_count_sequence(seq, regex=regex) for seq in seqs)

    return count


def merge_genomic_intervals(intervals):
    # type: (Iterable[Tuple[str, int, int]]) -> Iterable[Tuple[str, int, int]]
    """Merges overlapping genomic intervals.

    Parameters
    ----------
    intervals : List[tuple(str, int, int)]
        List of intervals to merge. Intervals are specified as tuples
        of (chomosome, start_position, end_position) values.

    Yields:
    tuple(str, int, int)
        Next merged interval.

    """

    # Group intervals by chromosome.
    grouped_intervals = itertools.groupby(
        sorted(intervals), operator.itemgetter(0))

    # Now yield merged intervals per chromosome.
    for chrom, grp in grouped_intervals:
        chrom_intervals = [interval[1:] for interval in grp]
        for low, high in merge_intervals(chrom_intervals, is_sorted=True):
            yield chrom, low, high


def merge_intervals(
        intervals,  # type: Iterable[Tuple[int, int]]
        is_sorted=False  # type: Optional[bool]
):  # type: (...) -> Iterable[Tuple[int, int]]
    """Merges overlapping intervals.

    Parameters
    ----------
    intervals : List[tuple(int, int)]
        List of intervals to merge. Intervals are specified as tuples
        of (start_position, end_position) values.
    is_sorted : bool
        Indicates if the intervals have already been sorted. Avoids
        re-sorting an already sorted list.

    Yields:
    tuple(int, int)
        Next merged interval.
    """

    if not is_sorted:
        intervals = sorted(intervals, key=operator.itemgetter(0))
    else:
        # Ensure intervals is a list.
        intervals = list(intervals)

    if not intervals:
        # No intervals to merge.
        return

    # Low and high represent the bounds of
    # the current run of merges.
    low, high = intervals[0]

    for iv in intervals[1:]:
        if iv[0] <= high:
            # New interval overlaps current run,
            # merge with the current run.
            high = max(high, iv[1])
        else:
            # Current run is over, yield accumulated
            # interval and start new run.
            yield low, high
            low, high = iv

    yield low, high



# -*- coding: utf-8 -*-
"""Provides functionality for creating and dealing with tabix-indexed files."""

# pylint: disable=wildcard-import,redefined-builtin,unused-wildcard-import
from __future__ import absolute_import, division, print_function
from builtins import *
# pylint: enable=wildcard-import,redefined-builtin,unused-wildcard-import

import contextlib
import csv
import itertools
import subprocess
from typing import Callable, Iterable, Any

from future.utils import native_str
import HTSeq as htseq
import pandas as pd
import pathlib2 as pathlib
import pysam

GTF_COLUMNS = [
    'seqname', 'source', 'feature', 'start', 'end', 'score', 'strand', 'frame',
    'attribute'
]

GTF_DTYPES = {
    'seqname': 'str',
    'source': 'category',
    'feature': 'category',
    'start': 'int',
    'end': 'int',
    'strand': 'category',
    'frame': 'category',
    'attribute': 'str'
}


class GtfIterator(object):
    """Iterator that iterates over records in a GTF file using pysam."""

    def __init__(self, gtf_path):
        # type: (pathlib.Path) -> None
        self._gtf_path = gtf_path

    @contextlib.contextmanager
    def _open_file(self):  # type: (...) -> pysam.TabixFile
        # Open gtf file.
        gtf_file = pysam.TabixFile(
            native_str(self._gtf_path), parser=pysam.asGTF())

        # Yield file object and ensure it is closed.
        try:
            yield gtf_file
        finally:
            gtf_file.close()

    @property
    def contigs(self):  # type: (...) -> List[str]
        """Contigs present in the gtf file."""

        with self._open_file() as gtf_file:
            return gtf_file.contigs

    def fetch(
            self,
            reference=None,  # type: str
            start=None,  # type: int
            end=None,  # type: int
            filters=None  # type: Iterable[Callable]
    ):  # type (...) -> Iterable[pysam.ctabixproxies.GTFProxy]
        """Fetches gtf records (gtf rows) from the gtf file."""

        with self._open_file() as gtf_file:
            # For some reason pysam does not fetch all records if reference
            # is None under Python 2.7. To fix this, here we simply chain all
            # the contig records into one iterable.
            if reference is None:
                contigs = gtf_file.contigs
                records = itertools.chain.from_iterable((gtf_file.fetch(
                    reference=ref, start=start, end=end) for ref in contigs))
            else:
                records = gtf_file.fetch(
                    reference=reference, start=start, end=end)

            # Filter records on additional filters.
            if filters is not None:
                for filter_func in filters:
                    records = (rec for rec in records if filter_func(rec))

            for record in records:
                yield record

    def fetch_genes(self, gene_ids=None, **kwargs):
        # type: (List[str], **Any) -> Iterable[pysam.ctabixproxies.GTFProxy]
        """Fetches gene records from the GTF file."""

        # Fetch gene records.
        gene_filter = lambda rec: rec.feature == 'gene'
        records = self.fetch(filters=[gene_filter], **kwargs)

        if gene_ids is not None:
            # Run through records.
            sought_ids = set(gene_ids)

            hits = {}
            for rec in records:
                if rec['gene_id'] in sought_ids:
                    # Record hits.
                    hits[rec['gene_id']] = rec

                    # Stop if we found everything.
                    if len(hits) == len(gene_ids):
                        break

            # Yield results, returning None for genes that weren't found.
            for gene_id in gene_ids:
                yield hits.get(gene_id, None)

        else:
            for record in records:
                yield record


def index_gtf(file_path, output_path, sort=True, force=True):
    # type: (pathlib.Path, pathlib.Path) -> None
    """Compresses and indexes a gtf file using bgzip and tabix."""

    # Sort file before compressing and indexing.
    if sort:
        sorted_path = _append_suffix(output_path, '.srt')
        sort_gtf(file_path, output_path=sorted_path)
    else:
        sorted_path = file_path

    # Gzip and index file.
    pysam.tabix_compress(
        native_str(sorted_path),
        filename_out=native_str(output_path),
        force=force)
    pysam.tabix_index(native_str(output_path), preset='gff', force=force)

    # Clean up sort temp file.
    if sort:
        sorted_path.unlink()


def sort_gtf(file_path, output_path):
    # type: (pathlib.Path, pathlib.Path) -> None
    """Sorts a gtf file by position, required for indexing by tabix."""

    try:
        # Try sorting using shell command.
        _sort_gtf_shell(file_path, output_path)
    except subprocess.CalledProcessError:
        # Fall back into in-memory sort if shell command fails.
        # TODO: Log warning.
        _sort_gtf_pandas(file_path, output_path)


def _sort_gtf_shell(file_path, output_path):
    """Sorts gtf file externally using grep/sort shell command."""

    with open(str(output_path), 'w') as out_file:
        cmd = '(grep ^"#" {0}; grep -v ^"#" {0} | sort -k1,1 -k4,4n)'
        subprocess.check_call(
            cmd.format(file_path), stdout=out_file, shell=True)


def _sort_gtf_pandas(file_path, output_path):
    """Sorts gtf file in-memory using pandas."""

    gtf_frame = read_gtf_frame(file_path)
    gtf_frame = sort_gtf_frame(gtf_frame)
    write_gtf_frame(gtf_frame, output_path)


def _append_suffix(path_obj, suffix):
    # type: (pathlib.Path, str) -> pathlib.Path
    if len(path_obj.suffixes) == 0:
        return path_obj.with_suffix(suffix)
    else:
        return path_obj.with_suffix(path_obj.suffixes[-1] + suffix)


def read_gtf_frame(file_path):
    """Reads an uncompressed GTF file into a pandas DataFrame."""

    gtf_frame = pd.read_csv(
        str(file_path),
        sep='\t',
        names=GTF_COLUMNS,
        dtype=GTF_DTYPES,
        comment='#')

    # Correct start to be zero-based.
    gtf_frame['start'] -= 1

    return gtf_frame


def write_gtf_frame(gtf_frame, file_path):
    """Writes gtf frame to a GTF file."""

    # Correct start to be one-based.
    gtf_frame = gtf_frame.assign(start=lambda df: df['start'] + 1)

    gtf_frame.to_csv(
        str(file_path),
        sep='\t',
        header=False,
        index=False,
        quoting=csv.QUOTE_NONE)


def sort_gtf_frame(gtf_frame):
    """Sorts gtf frame by position."""
    return gtf_frame.sort_values(['seqname', 'start'], ascending=True)


def flatten_gtf_frame(gtf_frame):
    """Flattens gtf frame to non-overlapping exonic parts."""

    # Extract gene_id and transcript_id.
    gtf_exons = gtf_frame.loc[gtf_frame['feature'] == 'exon'].copy()

    gtf_exons['gene_id'] = gtf_exons['attribute'].str.extract(
        r'gene_id "(\w+)"', expand=True)
    gtf_exons['transcript_id'] = gtf_exons['attribute'].str.extract(
        r'transcript_id "(\w+)"', expand=True)

    # Build exon array.
    exons = htseq.GenomicArrayOfSets('auto', stranded=True)
    for tup in gtf_exons.itertuples():
        interval = htseq.GenomicInterval(
            tup.seqname, tup.start, tup.end, strand=tup.strand)
        exons[interval] += (tup.gene_id, tup.transcript_id)

    # Extract exonic sections from exon array.
    gtf_frame = _gtf_frame_from_exon_array(exons)
    gtf_frame = sort_gtf_frame(gtf_frame)

    return gtf_frame


def _gtf_frame_from_exon_array(exons):
    """Returns exon intervals from given GenomicArrayOfSets object."""

    def _row_gen(exons):
        """Extracts rows from exon array."""

        for interval, attrs in exons.steps():
            # Skip empty intervals (no gene).
            if len(attrs) > 0:

                # Only use steps that overlap with exactly one gene.
                gene_ids = set(gene_id for gene_id, _ in attrs)
                if len(gene_ids) == 1:

                    # Return tuple containing position + gene_id.
                    yield (interval.chrom, interval.start, interval.end,
                           interval.strand, next(iter(gene_ids)))

    def _attribute_strs(gene_ids):
        """Returns attribute strs for a group of gene exons."""

        gene_id = gene_ids.iloc[0]
        fmt_str = (
            'gene_id "{}"'.format(gene_id) + '; exonic_part_number "{:03d}"')

        return [fmt_str.format(i) for i in range(1, len(gene_ids) + 1)]

    # Extract exonic sections from exon array.
    flat_exons = pd.DataFrame.from_records(
        _row_gen(exons),
        columns=['seqname', 'start', 'end', 'strand', 'gene_id'])

    # Add missing columns to create a gtf frame.
    flat_exons = flat_exons.sort_values(by=['gene_id', 'start', 'end'])
    attrs = flat_exons.groupby('gene_id')['gene_id'].transform(_attribute_strs)

    gtf_frame = flat_exons.assign(
        source='imfusion_flatten',
        feature='exonic_part',
        score='.',
        frame='.',
        attribute=attrs)
    gtf_frame = gtf_frame.reindex(columns=GTF_COLUMNS)

    return gtf_frame



import pytest

from imfusion.util import tabix


@pytest.fixture
def gtf_path():
    """Returns path to example gtf file containing Smn1."""
    return pytest.helpers.data_path(
        'Mus_musculus.GRCm38.76.Smn1.gtf', relative_to=__file__)


class TestReadGtfFrame(object):
    """Unit tests for the read_gtf_frame function."""

    def test_example(self, gtf_path):
        """Test example gtf containing Smn1."""

        gtf = tabix.read_gtf_frame(gtf_path)

        # Check shape.
        assert gtf.shape == (76, 9)
        assert list(gtf.columns) == tabix.GTF_COLUMNS

        # Check if start position = -1.
        assert gtf.iloc[0]['start'] == 100124851


class TestFlattenGtfFrame(object):
    """Unit tests for the flatten_gtf_frame function."""

    def test_example(self, gtf_path):
        """Test example gtf containing Smn1."""

        gtf = tabix.read_gtf_frame(gtf_path)
        gtf_flat = tabix.flatten_gtf_frame(gtf)

        # Check shape.
        assert gtf_flat.shape == (25, 9)
        assert list(gtf_flat.columns) == tabix.GTF_COLUMNS

        # Check positions.
        head = gtf_flat.head()
        assert list(head['start']) == [
            100124851, 100124857, 100124893, 100124923, 100125818]
        assert list(head['end']) == [
            100124857, 100124893, 100124923, 100124965, 100125931]

        # Check other values.
        first = gtf_flat.iloc[0]
        assert first.seqname == '13'
        assert first.source == 'imfusion_flatten'
        assert first.feature == 'exonic_part'
        assert first.score == '.'
        assert first.frame == '.'
        assert first.attribute == ('gene_id "ENSMUSG00000021645"; '
                                   'exonic_part_number "001"')

# class TestFlattenGtfFrame(object):
#     """Unit tests for flatten_intervals function."""

#     def test_smn1(self, test_intervals, flattened_test_intervals):
#         """Test Smn1 example."""
#         flattened = list(tabix.flatten_gtf_frame(test_intervals))
#         assert flattened == flattened_test_intervals

    # def test_empty(self):
    #     """Test empty example."""
    #     flattened = list(tabix._flatten_intervals([]))
    #     assert flattened == []




# -*- coding: utf-8 -*-
"""Implements a statistical models used for DE tests."""

# pylint: disable=wildcard-import,redefined-builtin,unused-wildcard-import
from __future__ import absolute_import, division, print_function
from builtins import *
# pylint: enable=wildcard-import,redefined-builtin,unused-wildcard-import

import numpy as np

try:
    from rpy2.robjects import pandas2ri
    from rpy2.robjects.packages import importr

    pandas2ri.activate()

    # Note robject_translations is no longer needed in Rpy2 2.6+,
    # but is kept for compatibility with older Rpy2 versions.
    r_mass = importr('MASS')
    r_stats = importr(
        'stats', robject_translations={'format_perc': '_format_perc'})
except ImportError:
    r_mass = None
    r_stats = None


class NegativeBinomial(object):
    """ Models a negative binomial distribution.

    NegativeBinomial class that wraps functionality from the R MASS
    and stats packages for fitting and evaluating Negative Binomials.
    Requires rpy2 to be installed.

    """

    def __init__(self, mu=None, size=None):
        self._check_rpy()

        self._mu = mu
        self._size = size

    @classmethod
    def fit(cls, data):
        """First the distribution using the given observations."""
        cls._check_rpy()

        fit = r_mass.fitdistr(data, densfun='negative binomial')
        fit = dict(zip(fit.names, [np.array(item) for item in fit]))
        mu, size = fit['estimate'][1], fit['estimate'][0]

        return cls(mu=mu, size=size)

    @staticmethod
    def _check_rpy():
        """Checks if Mass and Stats are available from rpy2."""
        if r_mass is None or r_stats is None:
            raise ValueError('Rpy2 must be installed to use the '
                             'NegativeBinomial distribution.')

    def pdf(self, x, log=False):
        """Returns the pdf of the distribution."""

        return np.array(
            r_stats.dnbinom(
                x=x, mu=self._mu, size=self._size, log=log))

    def cdf(self, q, log_p=False):
        """Returns the cdf of the distribution."""

        return self._pnbinom(
            q, mu=self._mu, size=self._size, lower_tail=True, log_p=log_p)

    def sf(self, q, log_p=False):
        """Returns the sf of the distribution."""

        return self._pnbinom(
            q, mu=self._mu, size=self._size, lower_tail=False, log_p=log_p)

    @staticmethod
    def _pnbinom(q, mu, size, **kwargs):
        p_val = r_stats.pnbinom(q, mu=mu, size=size, **kwargs)
        return p_val[0] if len(p_val) == 1 else np.array(p_val)

    def __repr__(self):
        return 'NegativeBinomial(mu={}, size={})'.format(self._size, self._mu)

    def __str__(self):
        return self.__repr__()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import setuptools

with open('README.rst') as readme_file:
    README = readme_file.read()

with open('HISTORY.rst') as history_file:
    HISTORY = history_file.read()

INSTALL_REQUIRES = [
    'future', 'pandas>=0.19.0', 'numexpr', 'pysam>=0.9.1', 'toolz', 'pyfaidx',
    'scipy', 'intervaltree', 'pathlib2', 'htseq>=0.7.2', 'matplotlib',
    'seaborn', 'typing; python_version < "3.5"'
]

EXTRAS_REQUIRE = {
    'de_single': ['rpy2'],
    'dev': [
        'sphinx', 'sphinx-autobuild', 'sphinx-rtd-theme', 'bumpversion',
        'pytest>=2.7', 'pytest-mock', 'pytest-helpers-namespace', 'pytest-cov',
        'python-coveralls', 'seaborn'
    ]
}

setuptools.setup(
    name='imfusion',
    version='0.3.0',
    description=('Tool for identifying transposon insertions in '
                 'Insertional Mutagenesis screens from gene-transposon '
                 'fusions using single- and paired-end RNA-sequencing data.'),
    long_description=README + '\n\n' + HISTORY,
    url='https://github.com/jrderuiter/im-fusion',
    author='Julian de Ruiter',
    author_email='julianderuiter@gmail.com',
    license='MIT license',
    packages=setuptools.find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'imfusion = imfusion.main.dispatch:main',
            'imfusion-build = imfusion.main.build:main',
            'imfusion-insertions = imfusion.main.insertions:main',
            'imfusion-ctg = imfusion.main.ctg:main',
            'imfusion-expression = imfusion.main.expression:main',
            'imfusion-merge = imfusion.main.merge:main'
        ]
    },
    install_requires=INSTALL_REQUIRES,
    extras_require=EXTRAS_REQUIRE,
    zip_safe=False,
    classifiers=[])



"""Dispatcher script that redirects calls to subcommands.

This script allows users to call subcommands from a single binary
(``imfusion build ...`` instead of ``imfusion-build`` by redirecting calls to
to the appropriate subcommands.

"""

import argparse
import subprocess
import sys

import imfusion

VALID_SUBCOMMANDS = {'build', 'insertions', 'expression', 'merge', 'ctg'}


def main():
    """Main function, parses the subcommand and executes the right script."""

    # Parse subcommand + extra arguments (which are 'unknown' args).
    parser = argparse.ArgumentParser()
    parser.add_argument('subcommand', choices=VALID_SUBCOMMANDS)
    parser.add_argument(
        '--version',
        action='version',
        version='IM-Fusion ' + imfusion.__version__)

    args, extra_args = parser.parse_known_args()

    # Pass arguments to subcommand.
    binary = 'imfusion-' + args.subcommand
    process = subprocess.Popen([binary] + extra_args)
    process.wait()

    # Copy processes returncode.
    sys.exit(process.returncode)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""Tests for imfusion.expression.test module."""

# pylint: disable=wildcard-import,redefined-builtin,unused-wildcard-import
from __future__ import absolute_import, division, print_function
from builtins import *
# pylint: enable=wildcard-import,redefined-builtin,unused-wildcard-import

import pytest

from imfusion.expression.counts import read_exon_counts
from imfusion.expression import test
from imfusion.model import Insertion

from imfusion.util.frozendict import frozendict

# pylint: disable=no-self-use,redefined-outer-name


@pytest.fixture
def insertions():
    """Example insertions in ENSMUSG00000051951 (5 exons)."""
    return [
        # After third exon.
        Insertion(
            id='1',
            seqname='1',
            position=3207327,
            strand=1,
            support_junction=1,
            support_spanning=1,
            support=2,
            metadata=frozendict({
                'gene_id': 'ENSMUSG00000051951',
                'sample': 'S1'})),

        # After first exon.
        Insertion(
            id='2',
            seqname='1',
            position=3214491,
            strand=-1,
            support_junction=1,
            support_spanning=1,
            support=2,
            metadata=frozendict({
                'gene_id': 'ENSMUSG00000051951',
                'sample': 'S2'}))
    ] # yapf: disable


@pytest.fixture
def exon_counts():
    """Example exon counts."""

    file_path = pytest.helpers.data_path(
        'exon_counts.txt', relative_to=__file__)
    return read_exon_counts(file_path)


class TestSplitCounts(object):
    """Tests for split_counts function."""

    def test_proper_example(self, insertions, exon_counts):
        """Tests example with two insertions, after exons 2 and 4."""

        before, after, dropped = test.split_counts(
            exon_counts, insertions, gene_id='ENSMUSG00000051951')

        assert len(before) == 1
        assert len(after) == 2
        assert dropped == set()

    def test_proper_example_df(self, insertions, exon_counts):
        """Tests same example, using dataframe input for insertions."""

        insertion_df = Insertion.to_frame(insertions)

        before, after, dropped = test.split_counts(
            exon_counts, insertion_df, gene_id='ENSMUSG00000051951')

        assert len(before) == 1
        assert len(after) == 2
        assert dropped == set()

    def test_before_gene(self, insertions, exon_counts):
        """Tests dropping sample with an insertion before the gene"""

        insertions[1] = insertions[1]._replace(position=3215652)

        before, after, dropped = test.split_counts(
            exon_counts, insertions, gene_id='ENSMUSG00000051951')

        assert len(before) == 3
        assert len(after) == 2
        assert dropped == {'S2'}

    def test_after_gene(self, insertions, exon_counts):
        """Tests dropping sample with an insertion after the gene"""

        insertions[0] = insertions[0]._replace(position=3205801)

        before, after, dropped = test.split_counts(
            exon_counts, insertions, gene_id='ENSMUSG00000051951')

        assert len(before) == 1
        assert len(after) == 4
        assert dropped == {'S1'}

    def test_in_exon(self, insertions, exon_counts):
        """Tests insertion in fourth exon."""

        insertions[0] = insertions[0]._replace(position=3207217)

        before, after, dropped = test.split_counts(
            exon_counts, insertions, gene_id='ENSMUSG00000051951')

        assert len(before) == 1
        assert len(after) == 1
        assert dropped == set()


@pytest.fixture
def test_insertions():
    """Example insertions for Trp53bp2 and Nf1."""

    file_path = pytest.helpers.data_path(
        'insertions.txt', relative_to=__file__)
    return list(Insertion.from_csv(file_path, sep='\t'))


@pytest.fixture
def test_exon_counts():
    """Example exon counts for Trp53bp2 and Nf1."""

    file_path = pytest.helpers.data_path(
        'exon_counts_test.txt', relative_to=__file__)
    return read_exon_counts(file_path)


class TestTestDeExon(object):
    """Tests for test_de function."""

    def test_pos_example_trp53bp2(self, test_insertions, test_exon_counts):
        """Tests positive example of DE in Trp53bp2 in the SB dataset."""

        result = test.test_de_exon(
            test_insertions, test_exon_counts, gene_id='ENSMUSG00000026510')

        assert result.p_value < 0.01
        assert result.direction == 1

    def test_pos_example_trp53bp2_df(self, test_insertions, test_exon_counts):
        """Tests positive example of DE in Trp53bp2 with dataframe input."""

        test_insertion_df = Insertion.to_frame(test_insertions)

        result = test.test_de_exon(
            test_insertion_df, test_exon_counts, gene_id='ENSMUSG00000026510')

        assert result.p_value < 0.01
        assert result.direction == 1

    def test_neg_example_nf1(self, test_insertions, test_exon_counts):
        """Tests negative example of DE in Nf1 in the SB dataset."""

        result = test.test_de_exon(
            test_insertions, test_exon_counts, gene_id='ENSMUSG00000020716')

        assert result.p_value > 0.05
        assert result.direction == 1

    def test_plot_boxplot(self, test_insertions, test_exon_counts):
        """Exercises plotting of boxplot."""

        result = test.test_de_exon(
            test_insertions, test_exon_counts, gene_id='ENSMUSG00000026510')
        result.plot_boxplot()

    def test_plot_sums(self, test_insertions, test_exon_counts):
        """Exercises plotting of sums."""

        result = test.test_de_exon(
            test_insertions, test_exon_counts, gene_id='ENSMUSG00000026510')
        result.plot_sums()



# -*- coding: utf-8 -*-
"""Tests for imfusion.merge module."""

# pylint: disable=wildcard-import,redefined-builtin,unused-wildcard-import
from __future__ import absolute_import, division, print_function
from builtins import *
# pylint: enable=wildcard-import,redefined-builtin,unused-wildcard-import

import pytest

from pathlib2 import Path

from imfusion import merge

# pylint: disable=no-self-use,redefined-outer-name


@pytest.fixture
def dir_paths():
    """Returns path to directory containing merge example data."""

    base_path = pytest.helpers.data_path('merge', relative_to=__file__)
    return [fp for fp in base_path.iterdir() if fp.is_dir()]


class TestMergeSamples(object):
    """Tests for merge_samples function."""

    def test_merging_insertions(self, dir_paths):
        """Tests merging of insertions."""
        # Merge insertions.
        insertions, _ = merge.merge_samples(dir_paths, with_expression=False)
        insertions = list(insertions)

        # Check result length.
        assert len(insertions) == 18

        # Check if sample names were used.
        assert insertions[0].metadata['sample'] == dir_paths[0].name
        assert insertions[0].id.startswith(dir_paths[0].name)

        assert insertions[9].metadata['sample'] == dir_paths[1].name
        assert insertions[9].id.startswith(dir_paths[1].name)

    def test_merge_ins_with_names(self, dir_paths):
        """Tests merging of insertions with sample names."""
        # Merge insertions.
        insertions, _ = merge.merge_samples(
            dir_paths, sample_names=['a', 'b'], with_expression=False)
        insertions = list(insertions)

        # Check result length.
        assert len(insertions) == 18

        # Check if sample names were used.
        assert insertions[0].metadata['sample'] == 'a'
        assert insertions[0].id.startswith('a')

        assert insertions[9].metadata['sample'] == 'b'
        assert insertions[9].id.startswith('b')

    def test_merge_ins_with_duplicates(self, dir_paths):
        """Tests merging of insertions with duplicate names."""

        with pytest.raises(ValueError):
            list(
                merge.merge_samples(
                    dir_paths, sample_names=['a', 'a'],
                    with_expression=False)[0]) # yapf: disable



# -*- coding: utf-8 -*-
"""Implements functions for merging datasets from multiple samples."""

# pylint: disable=wildcard-import,redefined-builtin,unused-wildcard-import
from __future__ import absolute_import, division, print_function
from builtins import *
# pylint: enable=wildcard-import,redefined-builtin,unused-wildcard-import

from typing import Iterable, Optional, Tuple
import itertools

import pathlib2 as pathlib

import pandas as pd
import toolz

from imfusion.expression import read_exon_counts
from imfusion.model import Insertion


def merge_samples(
        dir_paths,  # type: Iterable[pathlib.Path]
        sample_names=None,  # type: List[str]
        with_expression=True  # type: bool
):  # type: (...) -> Tuple[Iterable[Insertion], Optional[pd.DataFrame]]
    """Merges samples in dir_paths to a single set of insertions and counts.

    Parameters
    ----------
    dir_paths : List[pathlib.Path]
        Paths to the sample directories.
    samples : List[str]
        Names to use the merged samples.
    with_expression : bool
        Whether to also merge expression counts (assumes that expression.txt
        files have been generated using imfusion-expression).

    Returns
    -------
    Tuple[pandas.DataFrame, pandas.DataFrame]
        Two DataFrames respectively containing the merged insertions
        and the merged exon counts. If with_expression is False, the
        merged counts frame is returned as None.

    """

    # Use directory names as sample names if none given.
    if sample_names is None:
        sample_names = [dp.name for dp in dir_paths]

    # Check for duplicate names.
    if len(sample_names) != len(set(sample_names)):
        raise ValueError('Sample names contains duplicates')

    # Merge insertions.
    ins_paths = [dp / 'insertions.txt' for dp in dir_paths]
    ins_groups = (Insertion.from_csv(fp, sep='\t') for fp in ins_paths)
    merged_insertions = merge_insertions(ins_groups, sample_names)

    # Merge counts.
    if with_expression:
        count_paths = [dp / 'expression.txt' for dp in dir_paths]
        count_frames = (read_exon_counts(cp) for cp in count_paths)
        merged_counts = merge_exon_counts(count_frames, sample_names)
    else:
        merged_counts = None

    return merged_insertions, merged_counts


def merge_insertions(insertion_groups, sample_names):
    # type: (Iterable[Iterable[Insertion]], List[str]) -> Iterable[Insertion]
    """Merges insertions from different samples."""

    # Check for duplicate names.
    if len(sample_names) != len(set(sample_names)):
        raise ValueError('Sample names contains duplicates')

    # Merge renamed insertions into single generator.
    sample_ins = (_apply_sample_name(grp, sample)
                  for (grp, sample) in zip(insertion_groups, sample_names))
    merged = itertools.chain.from_iterable(sample_ins)

    # Yield result.
    for insertion in merged:
        yield insertion


def _apply_sample_name(insertions, sample_name):
    # type: (Iterable[Insertion], str) -> Iterable[Insertion]
    for ins in insertions:
        new_metadata = toolz.merge(ins.metadata, {'sample': sample_name})
        yield ins._replace(id='{}.{}'.format(sample_name, ins.id),
                           metadata=new_metadata) # yapf: disable


def merge_exon_counts(count_frames, sample_names):
    # type: (Iterable[pd.DataFrame], List[str]) -> pd.DataFrame
    """Merges expression count frames into a single frame."""

    # Check for duplicate names.
    if len(sample_names) != len(set(sample_names)):
        raise ValueError('Sample names contains duplicates')

    # Merge counts.
    merged = pd.concat(count_frames, axis=1)
    merged.columns = sample_names

    return merged



# -*- coding: utf-8 -*-
"""Tests for imfusion.model module."""

# pylint: disable=wildcard-import,redefined-builtin,unused-wildcard-import
from __future__ import absolute_import, division, print_function
from builtins import *
# pylint: enable=wildcard-import,redefined-builtin,unused-wildcard-import

import pytest

from imfusion.model import Fusion, TransposonFusion, Insertion
from imfusion.util.frozendict import frozendict

# pylint: disable=no-self-use,redefined-outer-name

# @pytest.fixture
# def merge_fusions():
#     junction_fusions = [Fusion(
#         seqname_a='1',
#         location_a=300,
#         strand_a=1,
#         seqname_b='T2onc',
#         location_b=420,
#         strand_b=1,
#         flank_a=52,
#         flank_b=62,
#         support_junction=2,
#         support_spanning=0)]

#     spanning_fusions = [Fusion(
#         seqname_a='1',
#         location_a=280,
#         strand_a=1,
#         seqname_b='T2onc',
#         location_b=445,
#         strand_b=1,
#         flank_a=0,
#         flank_b=0,
#         support_junction=0,
#         support_spanning=1)]

#     return junction_fusions, spanning_fusions


@pytest.fixture
def fusion():
    """Example fusion."""
    return Fusion(
        seqname_a='1',
        location_a=300,
        strand_a=1,
        seqname_b='T2onc',
        location_b=420,
        strand_b=1,
        flank_a=52,
        flank_b=62,
        support_junction=2,
        support_spanning=1)


# class TestToTransposonFusion(object):
#     def test_example(self, fusion):
#         expected = TransposonFusion(
#             seqname='1',
#             anchor_genome=300,
#             anchor_transposon=420,
#             strand_genome=1,
#             strand_transposon=1,
#             flank_genome=52,
#             flank_transposon=62,
#             support_junction=2,
#             support_spanning=1,
#             metadata={})

#         assert fusion.to_transposon_fusion('T2onc') == expected

#     def test_no_tranposon(self, fusion):
#         fusion = fusion._replace(seqname_b='15')
#         with pytest.raises(ValueError):
#             fusion.to_transposon_fusion('T2onc')

#     def test_no_genomic(self, fusion):
#         fusion = fusion._replace(seqname_a='T2onc')
#         with pytest.raises(ValueError):
#             fusion.to_transposon_fusion('T2onc')


@pytest.fixture
def tr_fusion():
    """Example transposon fusion."""
    return TransposonFusion(
        seqname='16',
        anchor_genome=52141095,
        anchor_transposon=1541,
        strand_genome=-1,
        strand_transposon=1,
        flank_genome=-78,
        flank_transposon=-76,
        support_junction=380,
        support_spanning=118,
        metadata=frozendict({
            'gene_strand': 1,
            'gene_name': 'Cblb',
            'feature_type': 'SA',
            'feature_strand': -1,
            'feature_name': 'En2SA'
        }))


class TestTransposonFusion(object):
    """Tests for TransposonFusion methods/properties."""

    def test_support(self, tr_fusion):
        """Tests support property."""
        expected = tr_fusion.support_junction + tr_fusion.support_spanning
        assert tr_fusion.support == expected

    def test_transposon_region(self, tr_fusion):
        """Tests transposon_region property."""
        assert tr_fusion.transposon_region == (1465, 1541)

    def test_genome_region(self, tr_fusion):
        """Tests genome_region property."""
        assert tr_fusion.genome_region == ('16', 52141017, 52141095)

    def test_from_fusion(self, fusion):
        """Tests from_fusion method."""

        tr_fusion = TransposonFusion.from_fusion(fusion, 'T2onc')

        expected = TransposonFusion(
            seqname='1',
            anchor_genome=300,
            anchor_transposon=420,
            strand_genome=1,
            strand_transposon=1,
            flank_genome=-52,
            flank_transposon=62,
            support_junction=2,
            support_spanning=1,
            metadata=frozendict({}))

        assert tr_fusion == expected
        assert isinstance(tr_fusion.metadata, frozendict)


class TestInsertion(object):
    """Tests methods/properties from Insertion class."""

    def test_from_transposon_fusion(self, tr_fusion):
        """Tests from_tranpsoson_fusion for example."""

        insertion = Insertion.from_transposon_fusion(tr_fusion)

        expected = Insertion(
            id=None,
            seqname='16',
            position=52141095,
            strand=-1,
            support_junction=380,
            support_spanning=118,
            support=498,
            metadata={
                'transposon_anchor': 1541,
                'orientation': 'antisense',
                'gene_name': 'Cblb',
                'feature_type': 'SA',
                'feature_strand': -1,
                'feature_name': 'En2SA',
                'gene_strand': 1
            })

        assert insertion == expected
        assert isinstance(insertion.metadata, frozendict)

    def test_from_transposon_fusion_id(self, tr_fusion):
        """Tests from_tranpsoson_fusion for example, with id."""

        insertion = Insertion.from_transposon_fusion(tr_fusion, id_='a')
        assert insertion.id == 'a'

    def test_from_transposon_fusions(self, tr_fusion):
        """Tests from_tranpsoson_fusions for example."""

        insertions = list(Insertion.from_transposon_fusions([tr_fusion]))

        expected = [
            Insertion(
                id=None,
                seqname='16',
                position=52141095,
                strand=-1,
                support_junction=380,
                support_spanning=118,
                support=498,
                metadata={
                    'transposon_anchor': 1541,
                    'orientation': 'antisense',
                    'gene_name': 'Cblb',
                    'feature_type': 'SA',
                    'feature_strand': -1,
                    'feature_name': 'En2SA',
                    'gene_strand': 1
                })
        ]

        assert expected == insertions

    def test_from_transposon_fusions_id(self, tr_fusion):
        """Tests from_tranpsoson_fusions for example, with id."""

        insertions = list(
            Insertion.from_transposon_fusions(
                [tr_fusion], id_fmt_str='INS_{}'))

        assert insertions[0].id == 'INS_1'



# from .annotate import annotate_fusions
# from .convert import convert_to_insertions
# from .filter import (filter_blacklist, filter_invalid_insertions,
#                      filter_unexpected_sites, filter_wrong_orientation)
